# ***************************************************************************
# *   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import unittest

import FreeCAD
import TestSketcherApp


class TestMirrored(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestMirrored")

    def testMirroredSketchCase(self):
        """
        Creates a unit cube cornered at the origin and mirrors it about the Y axis.
        This operation should create a rectangular prism with volume 2.0.
        """
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Rect = self.Doc.addObject("Sketcher::SketchObject", "Rect")
        self.Body.addObject(self.Rect)
        TestSketcherApp.CreateRectangleSketch(self.Rect, (0, 0), (1, 1))
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Pad.Profile = self.Rect
        self.Pad.Length = 1
        self.Body.addObject(self.Pad)
        self.Doc.recompute()
        self.Mirrored = self.Doc.addObject("PartDesign::Mirrored", "Mirrored")
        self.Mirrored.Originals = [self.Pad]
        self.Mirrored.MirrorPlane = (self.Rect, ["V_Axis"])
        self.Body.addObject(self.Mirrored)
        self.Doc.recompute()
        self.assertAlmostEqual(self.Mirrored.Shape.Volume, 2.0)

    def testMirroredPrimitiveCase(self):
        """
        Tests the same mirroring scenario as in the sketch case,
        but is designed to ensure the same end result occurs with
        a different base object.
        """
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Box = self.Doc.addObject("PartDesign::AdditiveBox", "Box")
        self.Box.Length = 1
        self.Box.Width = 1
        self.Box.Height = 1
        self.Body.addObject(self.Box)
        self.Doc.recompute()
        self.Mirrored = self.Doc.addObject("PartDesign::Mirrored", "Mirrored")
        self.Mirrored.Originals = [self.Box]
        self.Mirrored.MirrorPlane = (self.Doc.XY_Plane, [""])
        self.Body.addObject(self.Mirrored)
        self.Doc.recompute()
        self.assertAlmostEqual(self.Mirrored.Shape.Volume, 2.0)

    def testMirroredOffsetCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Rect = self.Doc.addObject("Sketcher::SketchObject", "Rect")
        self.Body.addObject(self.Rect)
        TestSketcherApp.CreateRectangleSketch(self.Rect, (0, 1), (1, 1))
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Pad.Profile = self.Rect
        self.Pad.Length = 1
        self.Body.addObject(self.Pad)
        self.Doc.recompute()
        self.Mirrored = self.Doc.addObject("PartDesign::Mirrored", "Mirrored")
        self.Mirrored.Originals = [self.Pad]
        self.Mirrored.MirrorPlane = (self.Rect, ["H_Axis"])
        self.Body.addObject(self.Mirrored)
        self.Doc.recompute()
        self.assertIn("Up-to-date", self.Mirrored.State)

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("PartDesignTestMirrored")
        # print ("omit closing document for debugging")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Base.Util as PathUtil
from typing import Dict, List, Any, Optional, Tuple
import tempfile
import os


def find_shape_object(doc: "FreeCAD.Document") -> Optional["FreeCAD.DocumentObject"]:
    """
    Find the primary object representing the shape in a document.

    Looks for PartDesign::Body, then Part::Feature. Falls back to the first
    object if no better candidate is found.

    Args:
        doc (FreeCAD.Document): The document to search within.

    Returns:
        Optional[FreeCAD.DocumentObject]: The found object or None.
    """
    obj = None
    # Prioritize Body
    for o in doc.Objects:
        if o.isDerivedFrom("PartDesign::Body"):
            return o
        # Keep track of the first Part::Feature found as a fallback
        if obj is None and o.isDerivedFrom("Part::Feature"):
            obj = o
    if obj:
        return obj
    # Fallback to the very first object if nothing else suitable found
    return doc.Objects[0] if doc.Objects else None


def get_unset_value_for(attribute_type: str):
    if attribute_type == "App::PropertyLength":
        return FreeCAD.Units.Quantity(0)
    elif attribute_type == "App::PropertyString":
        return ""
    elif attribute_type == "App::PropertyInteger":
        return 0
    return None


def get_object_properties(
    obj: "FreeCAD.DocumentObject",
    props: Optional[List[str]] = None,
    group: Optional[str] = None,
    exclude_groups: Optional[List[str]] = None,
) -> Dict[str, Tuple[Any, str]]:
    """
    Extract properties from a FreeCAD PropertyBag, including their types.

    Issues warnings for missing parameters but does not raise an error.

    Args:
        obj: The PropertyBag to extract properties from.
        props (List[str], optional): A list of property names to look for.
                                     If None, all properties in obj.PropertiesList are considered.
        group (str, optional): If provided, only properties belonging to this group are extracted.

    Returns:
        Dict[str, Tuple[Any, str]]: A dictionary mapping property names to a tuple
                                    (value, type_id). Values are FreeCAD native types.
                                    If a property is missing, its value will be None.
    """
    properties = {}
    for name in props or obj.PropertiesList:
        if group and not obj.getGroupOfProperty(name) == group:
            continue
        if exclude_groups and obj.getGroupOfProperty(name) in exclude_groups:
            continue
        if hasattr(obj, name):
            value = getattr(obj, name)
            type_id = obj.getTypeIdOfProperty(name)
            properties[name] = value, type_id
        else:
            # Log a warning if a parameter expected by the shape class is missing
            Path.Log.debug(
                f"Parameter '{name}' not found on object '{obj.Label}' "
                f"({obj.Name}). Default value will be used by the shape class."
            )
            properties[name] = None, "App::PropertyString"
    return properties


def update_shape_object_properties(
    obj: "FreeCAD.DocumentObject", parameters: Dict[str, Any]
) -> None:
    """
    Update properties of a FreeCAD PropertyBag based on a dictionary of parameters.

    Args:
        obj (FreeCAD.DocumentObject): The PropertyBag to update properties on.
        parameters (Dict[str, Any]): A dictionary of property names and values.
    """
    for name, value in parameters.items():
        if hasattr(obj, name):
            try:
                PathUtil.setProperty(obj, name, value)
            except Exception as e:
                Path.Log.warning(
                    f"Failed to set property '{name}' on object '{obj.Label}'"
                    f" ({obj.Name}) with value '{value}': {e}"
                )
        else:
            # This simply means that the Shape File doesn't have this property in it.
            Path.Log.debug(
                f"Property '{name}' not found on object '{obj.Label}' ({obj.Name}). Skipping."
            )


def get_doc_state() -> Any:
    """
    Used to make a "snapshot" of the current state of FreeCAD, to allow
    for restoring the ActiveDocument and selection state later.
    """
    doc_name = FreeCAD.ActiveDocument.Name if FreeCAD.ActiveDocument else None
    if FreeCAD.GuiUp:
        import FreeCADGui

        selection = FreeCADGui.Selection.getSelection()
    else:
        selection = []
    return doc_name, selection


def restore_doc_state(state):
    doc_name, selection = state
    if doc_name:
        FreeCAD.setActiveDocument(doc_name)
    if FreeCAD.GuiUp:
        import FreeCADGui

        for sel in selection:
            FreeCADGui.Selection.addSelection(doc_name, sel.Name)


class ShapeDocFromBytes:
    """
    Context manager to create and manage a temporary FreeCAD document,
    loading content from a byte string.
    """

    def __init__(self, content: bytes):
        self._content = content
        self._doc = None
        self._temp_file = None
        self._old_state = None

    def __enter__(self) -> "FreeCAD.Document":
        """Creates a new temporary FreeCAD document or loads cache if provided."""
        # Create a temporary file and write the cache content to it
        with tempfile.NamedTemporaryFile(suffix=".FCStd", delete=False) as tmp_file:
            tmp_file.write(self._content)
            self._temp_file = tmp_file.name

        # When we open a new document, FreeCAD loses the state, of the active
        # document (i.e. current selection), even if the newly opened document
        # is a hidden one.
        # So we need to restore the active document state at the end.
        self._old_state = get_doc_state()

        # Open the document from the temporary file
        # Use a specific name to avoid clashes if multiple docs are open
        # Open the document from the temporary file
        self._doc = FreeCAD.openDocument(self._temp_file, hidden=True)
        if not self._doc:
            raise RuntimeError(f"Failed to open document from {self._temp_file}")
        return self._doc

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Closes the temporary FreeCAD document and cleans up the temp file."""
        if self._doc:
            # Note that .closeDocument() is extremely slow; it takes
            # almost 400ms per document - much longer than opening!
            FreeCAD.closeDocument(self._doc.Name)
            self._doc = None

        # Restore the original active document
        restore_doc_state(self._old_state)

        # Clean up the temporary file if it was created
        if self._temp_file and os.path.exists(self._temp_file):
            try:
                os.remove(self._temp_file)
            except Exception as e:
                Path.Log.warning(f"Failed to remove temporary file {self._temp_file}: {e}")



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2015 Yorik van Havre <yorik@uncreated.net>
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

import os
import functools
import shutil
import tempfile
import threading
from typing import Dict, List

from PySideWrapper import QtGui, QtCore, QtWidgets

from addonmanager_workers_startup import (
    CreateAddonListWorker,
    CheckWorkbenchesForUpdatesWorker,
    GetBasicAddonStatsWorker,
    GetAddonScoreWorker,
    CheckForMissingDependenciesWorker,
)
from addonmanager_installer_gui import (
    AddonInstallerGUI,
    MacroInstallerGUI,
    AddonDependencyInstallerGUI,
)
from addonmanager_icon_utilities import get_icon_for_addon
from addonmanager_uninstaller_gui import AddonUninstallerGUI
from addonmanager_update_all_gui import UpdateAllGUI
import addonmanager_utilities as utils
import addonmanager_freecad_interface as fci
from composite_view import CompositeView
from Widgets.addonmanager_widget_global_buttons import WidgetGlobalButtonBar
from Widgets.addonmanager_widget_progress_bar import Progress
from Widgets.addonmanager_utility_dialogs import MessageDialog
from package_list import PackageListItemModel
from Addon import Addon, cycle_to_sub_addon, MissingDependencies
from addonmanager_python_deps_gui import (
    PythonPackageManagerGui,
)
from addonmanager_firstrun import FirstRunDialog
from addonmanager_connection_checker import ConnectionCheckerGUI

from addonmanager_metadata import MetadataReader

import NetworkManager

from AddonManagerOptions import AddonManagerOptions

translate = fci.translate


def QT_TRANSLATE_NOOP(_, txt):
    return txt


__title__ = "FreeCAD Addon Manager Module"
__author__ = "Yorik van Havre", "Jonathan Wiedemann", "Kurt Kremitzki", "Chris Hennes"
__url__ = "https://www.freecad.org"

"""
FreeCAD Addon Manager Module

Fetches various types of addons from a variety of sources. Built-in sources are:
* https://github.com/FreeCAD/FreeCAD-addons
* https://github.com/FreeCAD/FreeCAD-macros
* https://wiki.freecad.org/

Additional git sources may be configure via user preferences.

You need a working internet connection, and optionally git -- if git is not available, ZIP archives
are downloaded instead.
"""

#  \defgroup ADDONMANAGER AddonManager
#  \ingroup ADDONMANAGER
#  \brief The Addon Manager allows users to install workbenches and macros made by other users
#  @{

INSTANCE = None


class CommandAddonManager(QtCore.QObject):
    """The main Addon Manager class and FreeCAD command"""

    workers = [
        "create_addon_list_worker",
        "check_worker",
        "show_worker",
        "update_all_worker",
        "check_for_python_package_updates_worker",
        "get_basic_addon_stats_worker",
        "get_addon_score_worker",
        "check_missing_dependencies_worker",
    ]

    lock = threading.Lock()
    restart_required = False

    finished = QtCore.Signal()

    def __init__(self):
        super().__init__()

        QT_TRANSLATE_NOOP("QObject", "Addon Manager")
        if fci.GuiUp:
            fci.addPreferencePage(
                AddonManagerOptions,
                "Addon Manager",
            )

        self.item_model = None
        self.installer_gui = None
        self.composite_view = None
        self.button_bar = None

        self.dialog = None
        self.startup_sequence = []
        self.packages_with_updates = set()

        self.number_of_progress_regions = 0
        self.current_progress_region = 0

        self.check_worker = None
        self.check_for_python_package_updates_worker = None
        self.update_all_worker = None
        self.create_addon_list_worker = None
        self.get_addon_score_worker = None
        self.get_basic_addon_stats_worker = None
        self.check_missing_dependencies_worker = None

        self.manage_python_packages_dialog = None
        self.missing_dependency_installer = None

        # Set up the connection checker
        self.connection_checker = ConnectionCheckerGUI()
        self.connection_checker.connection_available.connect(self.launch)

        self.missing_dependencies = MissingDependencies()

        # Give other parts of the AM access to the current instance
        global INSTANCE
        INSTANCE = self

    def GetResources(self) -> Dict[str, str]:
        """FreeCAD-required function: get the core resource information for this Mod."""
        return {
            "Pixmap": "AddonManager",
            "MenuText": QT_TRANSLATE_NOOP("Std_AddonMgr", "&Addon Manager"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Std_AddonMgr",
                "Manages external workbenches, macros, and preference packs",
            ),
            "Group": "Tools",
        }

    def Activated(self) -> None:
        """FreeCAD-required function: called when the command is activated."""
        NetworkManager.InitializeNetworkManager()
        first_run_dialog = FirstRunDialog()
        if not first_run_dialog.exec():
            return
        self.connection_checker.start()

    def launch(self) -> None:
        """Shows the Addon Manager UI"""

        # create the dialog
        self.dialog = fci.loadUi(os.path.join(os.path.dirname(__file__), "AddonManager.ui"))
        self.dialog.setObjectName("AddonManager_Main_Window")
        # self.dialog.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)

        metadata = MetadataReader.from_file(os.path.join(os.path.dirname(__file__), "package.xml"))
        am_version = str(metadata.version)
        self.dialog.setWindowTitle(
            "{} {} {}".format(
                translate("AddonsInstaller", "Addon Manager"),
                translate("AddonsInstaller", "version"),
                am_version,
            )
        )

        # clean up the leftovers from previous runs
        self.packages_with_updates = set()
        self.startup_sequence = []
        self.cleanup_workers()

        # restore window geometry from the stored state
        w = fci.Preferences().get("WindowWidth")
        h = fci.Preferences().get("WindowHeight")
        self.composite_view = CompositeView(self.dialog)
        self.button_bar = WidgetGlobalButtonBar(self.dialog)

        self.button_bar.check_for_updates.hide()

        # Set up the listing of packages using the model-view-controller architecture
        self.item_model = PackageListItemModel()
        self.composite_view.setModel(self.item_model)
        self.dialog.layout().addWidget(self.composite_view)
        self.dialog.layout().addWidget(self.button_bar)

        # set nice icons to everything, by theme with fallback to FreeCAD icons
        icon_path = os.path.join(os.path.dirname(__file__), "Resources", "icons")
        self.dialog.setWindowIcon(QtGui.QIcon(os.path.join(icon_path, "addon_manager.svg")))

        # enable/disable stuff
        self.button_bar.update_all_addons.setEnabled(False)
        self.hide_progress_widgets()

        # connect slots
        self.dialog.rejected.connect(self.reject)
        self.dialog.accepted.connect(self.accept)
        self.button_bar.update_all_addons.clicked.connect(self.update_all)
        self.button_bar.close.clicked.connect(self.dialog.reject)
        self.button_bar.check_for_updates.clicked.connect(
            lambda: self.force_check_updates(standalone=True)
        )
        self.button_bar.python_dependencies.triggered.connect(self.show_python_updates_dialog)
        self.button_bar.addons_folder.triggered.connect(self.open_addons_folder)
        self.composite_view.package_list.stop_loading.connect(self.stop_update)
        self.composite_view.package_list.setEnabled(False)
        self.composite_view.execute.connect(self.execute_macro)
        self.composite_view.install.connect(self.update)
        self.composite_view.uninstall.connect(self.remove)
        self.composite_view.update.connect(self.update)
        self.composite_view.update_status.connect(self.status_updated)

        # center the dialog over the FreeCAD window if it exists
        self.dialog.resize(w, h)
        if fci.FreeCADGui:
            mw = fci.FreeCADGui.getMainWindow()
            self.dialog.move(
                mw.frameGeometry().topLeft() + mw.rect().center() - self.dialog.rect().center()
            )

        # begin populating the table in a set of sub-threads
        self.startup()

        # rock 'n roll!!!
        self.dialog.exec()

    def cleanup_workers(self) -> None:
        """Ensure that no workers are running by explicitly asking them to stop and waiting for
        them until they do"""
        for worker in self.workers:
            if hasattr(self, worker):
                thread = getattr(self, worker)
                if thread:
                    if not thread.isFinished():
                        thread.blockSignals(True)
                        thread.requestInterruption()
        for worker in self.workers:
            if hasattr(self, worker):
                thread = getattr(self, worker)
                if thread:
                    if not thread.isFinished():
                        finished = thread.wait(500)
                        if not finished:
                            fci.Console.PrintWarning(
                                translate(
                                    "AddonsInstaller",
                                    "Worker process {} is taking a long time to stop…",
                                ).format(worker)
                                + "\n"
                            )

    def accept(self) -> None:
        self.finished.emit()

    def reject(self) -> None:
        """called when the window has been closed"""

        # save window geometry for next use
        fci.Preferences().set("WindowWidth", self.dialog.width())
        fci.Preferences().set("WindowHeight", self.dialog.height())

        # ensure all threads are finished before closing
        ok_to_close = True
        self.startup_sequence = []
        for worker in self.workers:
            if hasattr(self, worker):
                thread = getattr(self, worker)
                if thread:
                    if not thread.isFinished():
                        thread.blockSignals(True)
                        thread.requestInterruption()
                        ok_to_close = False
        while not ok_to_close:
            ok_to_close = True
            for worker in self.workers:
                if hasattr(self, worker):
                    thread = getattr(self, worker)
                    if thread:
                        thread.wait(25)
                        if not thread.isFinished():
                            ok_to_close = False
            QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents)

        if self.restart_required:
            # display restart dialog
            m = QtWidgets.QMessageBox()
            m.setObjectName("AddonManager_RestartRequired")
            m.setWindowTitle(translate("AddonsInstaller", "Addon Manager"))
            icon_path = os.path.join(os.path.dirname(__file__), "Resources", "icons")
            m.setWindowIcon(QtGui.QIcon(os.path.join(icon_path, "addon_manager.svg")))
            m.setText(
                translate(
                    "AddonsInstaller",
                    "Restart FreeCAD for changes to take effect",
                )
            )
            m.setIcon(QtWidgets.QMessageBox.Icon.Warning)
            m.setStandardButtons(
                QtWidgets.QMessageBox.StandardButton.Ok
                | QtWidgets.QMessageBox.StandardButton.Cancel
            )
            m.setDefaultButton(QtWidgets.QMessageBox.StandardButton.Cancel)
            ok_btn = m.button(QtWidgets.QMessageBox.StandardButton.Ok)
            cancel_btn = m.button(QtWidgets.QMessageBox.StandardButton.Cancel)
            ok_btn.setText(translate("AddonsInstaller", "Restart Now"))
            cancel_btn.setText(translate("AddonsInstaller", "Restart Later"))
            ret = m.exec_()
            if ret == QtWidgets.QMessageBox.StandardButton.Ok:
                # restart FreeCAD after a delay to give time to this dialog to close
                QtCore.QTimer.singleShot(1000, utils.restart_freecad)

        self.finished.emit()

    def startup(self) -> None:
        """Downloads the available packages listings and populates the table"""

        # Each function in this list is expected to launch a thread and connect its completion
        # signal to self.do_next_startup_phase, or to shortcut to calling
        # self.do_next_startup_phase if it is not launching a worker
        self.startup_sequence = [
            self.populate_packages_table,
            self.activate_table_widgets,
            self.check_updates,
            self.check_missing_dependencies,
            self.fetch_addon_stats,
            self.fetch_addon_score,
            self.select_addon,
        ]
        self.number_of_progress_regions = len(self.startup_sequence)
        self.current_progress_region = 0
        self.do_next_startup_phase()

    def do_next_startup_phase(self) -> None:
        """Pop the top item in self.startup_sequence off the list and run it"""

        if len(self.startup_sequence) > 0:
            phase_runner = self.startup_sequence.pop(0)
            self.current_progress_region += 1
            self.update_progress_bar(translate("AddonsInstaller", "Continuing startup"), 0, 100)
            phase_runner()
        else:
            self.hide_progress_widgets()
            self.composite_view.package_list.item_filter.invalidateFilter()
            self.post_startup()

    def populate_packages_table(self) -> None:
        self.item_model.clear()

        self.create_addon_list_worker = CreateAddonListWorker()
        self.create_addon_list_worker.addon_repo.connect(self.add_addon_repo)
        self.update_progress_bar(translate("AddonsInstaller", "Creating addon list"), 10, 100)
        self.create_addon_list_worker.old_backups_found.connect(self.found_old_backups)
        self.create_addon_list_worker.finished.connect(self.do_next_startup_phase)  # Link to step 2
        self.create_addon_list_worker.progress_made.connect(self.update_progress_bar)
        self.create_addon_list_worker.start()

    def activate_table_widgets(self) -> None:
        self.composite_view.package_list.setEnabled(True)
        self.composite_view.package_list.ui.view_bar.search.setFocus()
        self.do_next_startup_phase()

    def on_package_updated(self, repo: Addon) -> None:
        """Called when the named package has either new metadata or a new icon (or both)"""

        with self.lock:
            repo.icon = get_icon_for_addon(repo, update=True)
            self.item_model.reload_item(repo)

    def select_addon(self) -> None:
        prefs = fci.Preferences()
        selection = prefs.get("SelectedAddon")
        if selection:
            self.composite_view.package_list.select_addon(selection)
            prefs.set("SelectedAddon", "")
        self.do_next_startup_phase()

    def check_updates(self) -> None:
        """checks every installed addon for available updates"""

        if not self.packages_with_updates:
            self.force_check_updates(standalone=False)
        else:
            self.do_next_startup_phase()

    def force_check_updates(self, standalone=False) -> None:
        if hasattr(self, "check_worker"):
            thread = self.check_worker
            if thread:
                if not thread.isFinished():
                    self.do_next_startup_phase()
                    return

        self.button_bar.update_all_addons.setText(
            translate("AddonsInstaller", "Checking for updates…")
        )
        self.packages_with_updates.clear()
        self.button_bar.update_all_addons.show()
        self.button_bar.check_for_updates.setDisabled(True)
        self.check_worker = CheckWorkbenchesForUpdatesWorker(self.item_model.repos)
        self.check_worker.finished.connect(self.do_next_startup_phase)
        self.check_worker.finished.connect(self.update_check_complete)
        self.check_worker.progress_made.connect(self.update_progress_bar)
        if standalone:
            self.current_progress_region = 1
            self.number_of_progress_regions = 1
        self.check_worker.update_status.connect(self.status_updated)
        self.check_worker.start()
        self.enable_updates(len(self.packages_with_updates))

    def status_updated(self, repo: Addon) -> None:
        self.item_model.reload_item(repo)
        if repo.status() == Addon.Status.UPDATE_AVAILABLE:
            self.packages_with_updates.add(repo)
            self.enable_updates(len(self.packages_with_updates))
        elif repo.status() == Addon.Status.PENDING_RESTART:
            self.restart_required = True

    def enable_updates(self, number_of_updates: int) -> None:
        """enables the update button"""

        if number_of_updates:
            self.button_bar.set_number_of_available_updates(number_of_updates)
        elif (
            hasattr(self, "check_worker")
            and self.check_worker is not None
            and self.check_worker.isRunning()
        ):
            self.button_bar.update_all_addons.setText(
                translate("AddonsInstaller", "Checking for updates…")
            )
        else:
            self.button_bar.set_number_of_available_updates(0)

    def update_check_complete(self) -> None:
        self.enable_updates(len(self.packages_with_updates))
        self.button_bar.check_for_updates.setEnabled(True)

    def check_missing_dependencies(self) -> None:
        """See if we have any missing dependencies"""
        self.check_missing_dependencies_worker = CheckForMissingDependenciesWorker(
            self.item_model.repos
        )
        self.update_progress_bar(translate("AddonsInstaller", "Checking dependencies"), 0, 100)
        self.check_missing_dependencies_worker.progress.connect(self.update_progress_bar)
        self.check_missing_dependencies_worker.finished.connect(self.do_next_startup_phase)
        self.check_missing_dependencies_worker.start()

    def show_python_updates_dialog(self) -> None:
        if not self.manage_python_packages_dialog:
            self.manage_python_packages_dialog = PythonPackageManagerGui(self.item_model.repos)
        self.manage_python_packages_dialog.show()

    def fetch_addon_stats(self) -> None:
        """Fetch the Addon Stats JSON data from a URL"""
        self.update_progress_bar(translate("AddonsInstaller", "Fetching addon stats"), 0, 100)
        url = fci.Preferences().get("AddonsStatsURL")
        if url and url != "NONE":
            self.get_basic_addon_stats_worker = GetBasicAddonStatsWorker(
                url, self.item_model.repos, self.dialog
            )
            self.get_basic_addon_stats_worker.finished.connect(self.do_next_startup_phase)
            self.get_basic_addon_stats_worker.update_addon_stats.connect(self.update_addon_stats)
            self.get_basic_addon_stats_worker.start()
        else:
            self.do_next_startup_phase()

    def update_addon_stats(self, addon: Addon):
        self.item_model.reload_item(addon)

    def fetch_addon_score(self) -> None:
        """Fetch the Addon score JSON data from a URL"""
        self.update_progress_bar(translate("AddonsInstaller", "Fetching addon score"), 0, 100)
        prefs = fci.Preferences()
        url = prefs.get("AddonsScoreURL")
        if url and url != "NONE":
            self.get_addon_score_worker = GetAddonScoreWorker(
                url, self.item_model.repos, self.dialog
            )
            self.get_addon_score_worker.finished.connect(self.score_fetched_successfully)
            self.get_addon_score_worker.finished.connect(self.do_next_startup_phase)
            self.get_addon_score_worker.update_addon_score.connect(self.update_addon_score)
            self.get_addon_score_worker.start()
        else:
            self.composite_view.package_list.ui.view_bar.set_rankings_available(False)
            self.do_next_startup_phase()

    def update_addon_score(self, addon: Addon):
        self.item_model.reload_item(addon)

    def score_fetched_successfully(self):
        self.composite_view.package_list.ui.view_bar.set_rankings_available(True)

    def add_addon_repo(self, addon_repo: Addon) -> None:
        """adds a workbench to the list"""

        if addon_repo.icon is None or addon_repo.icon.isNull():
            addon_repo.icon = get_icon_for_addon(addon_repo)
        for repo in self.item_model.repos:
            if repo.name == addon_repo.name:
                # self.item_model.reload_item(repo) # If we want to have later additions superseded
                # earlier
                return
        self.item_model.append_item(addon_repo)

    def append_to_repos_list(self, repo: Addon) -> None:
        """this function allows threads to update the main list of workbenches"""
        self.item_model.append_item(repo)

    def update(self, repo: Addon) -> None:
        try:
            self.prep_for_install(repo)
        except OSError as e:
            fci.Console.PrintError(e)
            fci.Console.PrintError("\n\nInstallation cancelled: out of disk space?\n")
            return
        self.launch_installer_gui(repo)

    def mark_repo_update_available(self, repo: Addon, available: bool) -> None:
        if available:
            repo.set_status(Addon.Status.UPDATE_AVAILABLE)
        else:
            repo.set_status(Addon.Status.NO_UPDATE_AVAILABLE)
        self.item_model.reload_item(repo)
        self.composite_view.package_details_controller.show_addon(repo)

    def prep_for_install(self, installing_addon: Addon):
        """To prepare for installing an addon, we need to see if this is the current active branch:
        if it is not, then we need to cycle the addon's attached to this addon ID, making this one
        active. We'll also remove any existing addon installed with this ID, making a backup
        so that we can recover from a failed update/branch switch."""

        for catalog_addon in self.item_model.repos:
            if catalog_addon.name == installing_addon.name:
                if catalog_addon != installing_addon:
                    cycle_to_sub_addon(catalog_addon, installing_addon, self.item_model)
                break
        make_backup(installing_addon)

    def launch_installer_gui(self, addon: Addon) -> None:
        if self.installer_gui is not None:
            fci.Console.PrintError(
                translate(
                    "AddonsInstaller",
                    "Cannot launch a new installer until the previous one has finished",
                )
            )
            return
        if addon.macro is not None:
            self.installer_gui = MacroInstallerGUI(addon)
        else:
            self.installer_gui = AddonInstallerGUI(addon, self.item_model.repos)
        self.installer_gui.success.connect(self.on_package_status_changed)
        self.installer_gui.success.connect(cleanup_pre_installation_backup)
        self.installer_gui.finished.connect(self.cleanup_installer)
        self.installer_gui.run()  # Does not block

    def cleanup_installer(self) -> None:
        QtCore.QTimer.singleShot(500, self.no_really_clean_up_the_installer)

    def no_really_clean_up_the_installer(self) -> None:
        self.installer_gui = None

    def update_all(self) -> None:
        """Asynchronously apply all available updates: individual failures are noted, but do not
        stop other updates"""

        if self.installer_gui is not None:
            fci.Console.PrintError(
                translate(
                    "AddonsInstaller",
                    "Cannot launch a new installer until the previous one has finished",
                )
            )
            return

        self.installer_gui = UpdateAllGUI(self.item_model.repos)
        self.installer_gui.addon_updated.connect(self.on_package_status_changed)
        self.installer_gui.finished.connect(self.cleanup_installer)
        self.installer_gui.run()  # Does not block

    def hide_progress_widgets(self) -> None:
        """hides the progress bar and related widgets"""
        self.composite_view.package_list.set_loading(False)

    def show_progress_widgets(self) -> None:
        self.composite_view.package_list.set_loading(True)

    def update_progress_bar(self, message: str, current_value: int, max_value: int) -> None:
        """Update the progress bar, showing it if it's hidden"""

        max_value = max_value if max_value > 0 else 1

        if current_value < 0:
            current_value = 0
        elif current_value > max_value:
            current_value = max_value

        self.show_progress_widgets()

        progress = Progress(
            status_text=message,
            number_of_tasks=self.number_of_progress_regions,
            current_task=self.current_progress_region - 1,
            current_task_progress=current_value / max_value,
        )
        self.composite_view.package_list.update_loading_progress(progress)

    def post_startup(self) -> None:
        """This is called after the startup sequence has completed"""
        if self.check_missing_dependencies_worker:
            deps: MissingDependencies = self.check_missing_dependencies_worker.missing_dependencies
            if deps.wbs or deps.external_addons or deps.python_requires:
                ignored_deps_string = fci.Preferences().get("ignored_missing_deps")
                ignored_deps = ignored_deps_string.split(";")

                proceed = False
                all_deps = set()
                all_deps.update(deps.wbs)
                all_deps.update(deps.external_addons)
                all_deps.update(deps.python_requires)
                for dep in all_deps:
                    if dep not in ignored_deps:
                        proceed = True
                        break

                if proceed:
                    self.missing_dependency_installer = AddonDependencyInstallerGUI([], deps)
                    self.missing_dependency_installer.dependency_dialog.label.setText(
                        translate(
                            "AddonsInstaller",
                            "Some installed addons are missing dependencies. Would you like to install them now?",
                        )
                    )
                    self.missing_dependency_installer.dependency_dialog.buttonBox.button(
                        QtWidgets.QDialogButtonBox.Ignore
                    ).clicked.connect(self.ignore_missing_dependencies)
                    self.missing_dependency_installer.run()

    def ignore_missing_dependencies(self):
        old_deps_string = fci.Preferences().get("ignored_missing_deps")
        old_deps = set(old_deps_string.split(";"))
        deps = self.check_missing_dependencies_worker.missing_dependencies
        new_deps = old_deps.union(deps.wbs)
        new_deps = new_deps.union(deps.external_addons)
        new_deps = new_deps.union(deps.python_requires)
        new_deps_string = ";".join(new_deps)
        fci.Preferences().set("ignored_missing_deps", new_deps_string)

    def stop_update(self) -> None:
        self.cleanup_workers()
        self.hide_progress_widgets()

    def on_package_status_changed(self, repo: Addon) -> None:
        if repo.status() == Addon.Status.PENDING_RESTART:
            self.restart_required = True
        self.item_model.reload_item(repo)
        self.composite_view.package_details_controller.show_addon(repo)
        if repo in self.packages_with_updates:
            self.packages_with_updates.remove(repo)
            self.enable_updates(len(self.packages_with_updates))

    def execute_macro(self, repo: Addon) -> None:
        """executes a selected macro"""

        if not fci.FreeCADGui:
            fci.Console.PrintError("Cannot execute a FreeCAD Macro outside FreeCAD")
            return

        macro = repo.macro
        if not macro or not macro.code:
            return

        if macro.is_installed():
            macro_path = os.path.join(fci.DataPaths().macro_dir, macro.filename)
            fci.FreeCADGui.open(str(macro_path))
            self.dialog.hide()
            fci.FreeCADGui.SendMsgToActiveView("Run")
        else:
            with tempfile.TemporaryDirectory() as temp_dir:
                temp_install_succeeded = macro.install(temp_dir)
                if not temp_install_succeeded:
                    fci.Console.PrintError(
                        translate("AddonsInstaller", "Temporary installation of macro failed")
                    )
                    return
                macro_path = os.path.join(temp_dir, macro.filename)
                fci.FreeCADGui.open(str(macro_path))
                self.dialog.hide()
                fci.FreeCADGui.SendMsgToActiveView("Run")

    def remove(self, addon: Addon) -> None:
        """Remove this addon."""
        if self.installer_gui is not None:
            fci.Console.PrintError(
                translate(
                    "AddonsInstaller",
                    "Cannot launch a new installer until the previous one has finished",
                )
            )
            return
        self.installer_gui = AddonUninstallerGUI(addon)
        self.installer_gui.finished.connect(self.cleanup_installer)
        self.installer_gui.finished.connect(
            functools.partial(self.on_package_status_changed, addon)
        )
        self.installer_gui.run()  # Does not block

    @staticmethod
    def open_addons_folder():
        addons_folder = fci.DataPaths().mod_dir
        QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(addons_folder))
        return

    @staticmethod
    def found_old_backups(backups: List[str]):
        handling = fci.Preferences().get("old_backup_handling")
        if handling.lower().strip() == "never":
            return
        if handling.lower().strip() == "always":
            delete_old_backups(backups)
            return

        backup_string = (
            translate(
                "AddonsInstaller",
                f"The following auto-generated backups were found in your Mod directory:",
            )
            + "\n"
        )
        for backup in backups:
            backup_string += "• " + str(backup) + "\n"
        backup_string += translate("AddonsInstaller", "Delete them now?")
        dlg = QtWidgets.QMessageBox()
        dlg.setObjectName("AddonManager_FoundOldBackups")
        dlg.setText(backup_string)
        dlg.setStandardButtons(
            QtWidgets.QMessageBox.YesToAll
            | QtWidgets.QMessageBox.Yes
            | QtWidgets.QMessageBox.No
            | QtWidgets.QMessageBox.NoToAll
        )
        dlg.setDefaultButton(QtWidgets.QMessageBox.No)
        dlg.button(QtWidgets.QMessageBox.YesToAll).setText(
            translate("AddonsInstaller", "Always", "'Always' delete old backups")
        )
        dlg.button(QtWidgets.QMessageBox.NoToAll).setText(
            translate("AddonsInstaller", "Never", "'Never' delete old backups")
        )
        result = dlg.exec_()

        if result == QtWidgets.QMessageBox.NoToAll:
            fci.Preferences().set("old_backup_handling", "never")
            return
        if result == QtWidgets.QMessageBox.No:
            return
        if result == QtWidgets.QMessageBox.YesToAll:
            fci.Preferences().set("old_backup_handling", "always")
        delete_old_backups(backups)


# Some utility functions


def make_backup(addon: Addon) -> None:
    """Make a backup of the addon's current installation directory, so that we can recover
    from a failed update/branch switch."""
    original = str(os.path.join(fci.DataPaths().mod_dir, addon.name))
    if os.path.exists(original):
        shutil.copytree(original, original + ".pre_update_backup", dirs_exist_ok=True)


def cleanup_pre_installation_backup(addon: Addon) -> None:
    """Remove the backup of the addon's current installation directory"""
    original = str(os.path.join(fci.DataPaths().mod_dir, addon.name))
    if os.path.exists(original):
        shutil.rmtree(original + ".pre_update_backup", ignore_errors=True)


def revert_to_backup(addon: Addon) -> None:
    """Revert to the backup of the addon's current installation directory"""
    original = str(os.path.join(fci.DataPaths().mod_dir, addon.name))
    if os.path.exists(original + ".pre_update_backup"):
        shutil.rmtree(original, ignore_errors=True)
        shutil.copytree(original + ".pre_update_backup", original, dirs_exist_ok=True)


def delete_old_backups(backups) -> None:
    """Delete old backups found in the Mod directory."""
    for backup in backups:
        full_path = str(os.path.join(fci.DataPaths().mod_dir, backup))
        if os.path.exists(full_path):
            success = utils.rmdir(full_path)
            if not success:
                fci.Console.PrintError(f"Failed to delete {full_path}\n")


# @}



# ***************************************************************************
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Collection of natural constants for the Fem module.

This module contains natural constants for the Fem module.
All constants are in SI units.
"""


__title__ = "FEM collection of natural constants"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


def gravity():
    # https://en.wikipedia.org/wiki/Gravitational_acceleration
    return "9.80665 m/s^2"


def stefan_boltzmann():
    # https://en.wikipedia.org/wiki/Stefan-Boltzmann_constant
    return "5.67037e-8 W/(m^2*K^4)"


def vacuum_permeability():
    # https://en.wikipedia.org/wiki/Vacuum_permeability
    return "1.256637e-6 N/A^2"


def vacuum_permittivity():
    # https://forum.freecad.org/viewtopic.php?f=18&p=400959#p400959
    # https://en.wikipedia.org/wiki/Permittivity#Vacuum_permittivity
    return "8.85419e-12 F/m"


def boltzmann_constant():
    # https://en.wikipedia.org/wiki/Boltzmann_constant
    return "1.38065e-23 J/K"


"""
from FreeCAD import Units
from femtools import constants
Units.Quantity(constants.gravity()).getValueAs("mm/s^2")

"""

# TODO: a unit test to be sure these values are returned!



# ***************************************************************************
# *   Copyright (c) 2013 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM material task panel for the document object"
__author__ = "Juergen Riegel, Bernd Hahnebach, Qingfeng Xia"
__url__ = "https://www.freecad.org"

## @package task_material_common
#  \ingroup FEM
#  \brief FreeCAD FEM _ViewProviderFemMaterial
#  \brief task panel for common material object

from PySide import QtCore

import FreeCAD
import FreeCADGui
from FreeCAD import Units
import Materials
import MatGui

from femguiutils import selection_widgets
from . import base_femtaskpanel


class _TaskPanel(base_femtaskpanel._BaseTaskPanel):
    """
    The editmode TaskPanel for FemMaterial objects
    """

    def __init__(self, obj):
        super().__init__(obj)

        # FreeCAD material dictionary of current material
        self.material = self.obj.Material
        self.uuid = self.obj.UUID
        self.material_manager = Materials.MaterialManager()

        # parameter widget
        self.parameterWidget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/Material.ui"
        )
        self.material_tree = MatGui.MaterialTreeWidget(self.parameterWidget.wgt_material_tree)
        self.material_tree.expanded = False
        self.material_tree.IncludeEmptyFolders = False
        self.material_tree.IncludeEmptyLibraries = False

        QtCore.QObject.connect(
            self.parameterWidget.chbu_allow_edit,
            QtCore.SIGNAL("clicked()"),
            self.toggleInputFieldsReadOnly,
        )
        # basic properties must be provided
        QtCore.QObject.connect(
            self.parameterWidget.qsb_density,
            QtCore.SIGNAL("editingFinished()"),
            self.density_changed,
        )
        # mechanical properties
        QtCore.QObject.connect(
            self.parameterWidget.qsb_young_modulus,
            QtCore.SIGNAL("editingFinished()"),
            self.ym_changed,
        )
        QtCore.QObject.connect(
            self.parameterWidget.qsb_poisson_ratio,
            QtCore.SIGNAL("editingFinished()"),
            self.pr_changed,
        )
        # thermal properties
        QtCore.QObject.connect(
            self.parameterWidget.qsb_thermal_conductivity,
            QtCore.SIGNAL("editingFinished()"),
            self.tc_changed,
        )
        QtCore.QObject.connect(
            self.parameterWidget.qsb_expansion_coefficient,
            QtCore.SIGNAL("editingFinished()"),
            self.tec_changed,
        )
        QtCore.QObject.connect(
            self.parameterWidget.qsb_expansion_reference_temperature,
            QtCore.SIGNAL("editingFinished()"),
            self.tec_changed,
        )
        QtCore.QObject.connect(
            self.parameterWidget.qsb_specific_heat,
            QtCore.SIGNAL("editingFinished()"),
            self.sh_changed,
        )
        # fluidic properties, only volumetric thermal expansion coeff makes sense
        QtCore.QObject.connect(
            self.parameterWidget.qsb_kinematic_viscosity,
            QtCore.SIGNAL("editingFinished()"),
            self.kinematic_viscosity_changed,
        )
        QtCore.QObject.connect(
            self.parameterWidget.wgt_material_tree,
            QtCore.SIGNAL("onMaterial(QString)"),
            self.set_from_editor,
        )

        # init all parameter input files with read only
        self.parameterWidget.chbu_allow_edit.setCheckState(QtCore.Qt.CheckState.Unchecked)
        self.toggleInputFieldsReadOnly()

        # hide some groupBox according to material category
        if self.obj.Category == "Fluid":
            self.filter_models(self.obj.Category)
            self.parameterWidget.groupBox_mechanical.setVisible(0)
        else:
            self.parameterWidget.groupBox_fluidic.setVisible(0)

        # geometry selection widget
        self.selectionWidget = selection_widgets.GeometryElementsSelection(
            obj.References, ["Solid", "Face", "Edge"], False, True
        )  # start with Solid in list!

        # form made from param and selection widget
        self.form = [self.parameterWidget, self.selectionWidget]

        # check references, has to be after initialisation of selectionWidget
        self.material_tree.UUID = self.uuid
        self.set_mat_params_in_input_fields(self.material)

        self.selectionWidget.has_equal_references_shape_types()

    # leave task panel ***************************************************************************
    def accept(self):
        if self.selectionWidget.has_equal_references_shape_types():
            self.obj.Material = self.material
            self.obj.UUID = self.uuid
            self.obj.References = self.selectionWidget.references

        self.selectionWidget.finish_selection()
        return super().accept()

    def reject(self):
        self.selectionWidget.finish_selection()
        return super().reject()

    # choose material ****************************************************************************
    def filter_models(self, category):
        material_filter = Materials.MaterialFilter()
        models = set(Materials.ModelManager().Models.keys())
        uuids = Materials.UUIDs()
        if category == "Fluid":
            material_filter.RequiredModels = [uuids.Fluid]
            self.material_tree.setFilter(material_filter)

    def toggleInputFieldsReadOnly(self):
        if self.parameterWidget.chbu_allow_edit.isChecked():
            self.parameterWidget.qsb_density.setDisabled(False)
            self.parameterWidget.qsb_young_modulus.setDisabled(False)
            self.parameterWidget.qsb_poisson_ratio.setDisabled(False)
            self.parameterWidget.qsb_thermal_conductivity.setDisabled(False)
            self.parameterWidget.qsb_expansion_coefficient.setDisabled(False)
            self.parameterWidget.qsb_expansion_reference_temperature.setDisabled(False)
            self.parameterWidget.qsb_specific_heat.setDisabled(False)
            self.parameterWidget.qsb_kinematic_viscosity.setDisabled(False)
            self.parameterWidget.wgt_material_tree.setEnabled(False)
            self.uuid = ""
            self.mat_from_input_fields()
        else:
            self.parameterWidget.qsb_density.setDisabled(True)
            self.parameterWidget.qsb_young_modulus.setDisabled(True)
            self.parameterWidget.qsb_poisson_ratio.setDisabled(True)
            self.parameterWidget.qsb_thermal_conductivity.setDisabled(True)
            self.parameterWidget.qsb_expansion_coefficient.setDisabled(True)
            self.parameterWidget.qsb_expansion_reference_temperature.setDisabled(True)
            self.parameterWidget.qsb_specific_heat.setDisabled(True)
            self.parameterWidget.qsb_kinematic_viscosity.setDisabled(True)
            self.parameterWidget.wgt_material_tree.setEnabled(True)
            self.set_from_editor(self.material_tree.UUID)

    # material parameter input fields ************************************************************
    # mechanical input fields
    def ym_changed(self):
        if self.parameterWidget.chbu_allow_edit.isChecked():
            self.material["YoungsModulus"] = self.parameterWidget.qsb_young_modulus.property(
                "value"
            ).UserString

    def density_changed(self):
        if self.parameterWidget.chbu_allow_edit.isChecked():
            self.material["Density"] = self.parameterWidget.qsb_density.property("value").UserString

    def pr_changed(self):
        if self.parameterWidget.chbu_allow_edit.isChecked():
            self.material["PoissonRatio"] = str(
                self.parameterWidget.qsb_poisson_ratio.property("value").Value
            )

    # thermal input fields
    def tc_changed(self):
        if self.parameterWidget.chbu_allow_edit.isChecked():
            self.material["ThermalConductivity"] = (
                self.parameterWidget.qsb_thermal_conductivity.property("value").UserString
            )

    def tec_changed(self):
        if self.parameterWidget.chbu_allow_edit.isChecked():
            self.material["ThermalExpansionCoefficient"] = (
                self.parameterWidget.qsb_expansion_coefficient.property("value").UserString
            )

    def tec_changed(self):
        if self.parameterWidget.chbu_allow_edit.isChecked():
            self.material["ThermalExpansionReferenceTemperature"] = (
                self.parameterWidget.qsb_expansion_reference_temperature.property(
                    "value"
                ).UserString
            )

    def sh_changed(self):
        if self.parameterWidget.chbu_allow_edit.isChecked():
            self.material["SpecificHeat"] = self.parameterWidget.qsb_specific_heat.property(
                "value"
            ).UserString

    # fluidic input fields
    def kinematic_viscosity_changed(self):
        if self.parameterWidget.chbu_allow_edit.isChecked():
            self.material["KinematicViscosity"] = (
                self.parameterWidget.qsb_kinematic_viscosity.property("value").UserString
            )

    def set_mat_params_in_input_fields(self, matmap):
        if "YoungsModulus" in matmap:
            self.parameterWidget.qsb_young_modulus.setProperty(
                "value", Units.Quantity(matmap["YoungsModulus"])
            )
        else:
            self.parameterWidget.qsb_young_modulus.setProperty("rawValue", 0.0)

        if "PoissonRatio" in matmap:
            self.parameterWidget.qsb_poisson_ratio.setProperty(
                "value", Units.Quantity(matmap["PoissonRatio"])
            )
        else:
            self.parameterWidget.qsb_poisson_ratio.setProperty("rawValue", 0.0)
        # Fluidic properties
        if "KinematicViscosity" in matmap:
            self.parameterWidget.qsb_kinematic_viscosity.setProperty(
                "value", Units.Quantity(matmap["KinematicViscosity"])
            )
        else:
            self.parameterWidget.qsb_kinematic_viscosity.setProperty("rawValue", 0.0)

        if "Density" in matmap:
            self.parameterWidget.qsb_density.setProperty("value", Units.Quantity(matmap["Density"]))
        else:
            self.parameterWidget.qsb_density.setProperty("rawValue", 0.0)
        # thermal properties
        if "ThermalConductivity" in matmap:
            self.parameterWidget.qsb_thermal_conductivity.setProperty(
                "value", Units.Quantity(matmap["ThermalConductivity"])
            )
        else:
            self.parameterWidget.qsb_thermal_conductivity.setProperty("rawValue", 0.0)

        if "ThermalExpansionCoefficient" in matmap:
            v = Units.Quantity(matmap["ThermalExpansionCoefficient"])
            v.Format = {"Precision": 3}
            self.parameterWidget.qsb_expansion_coefficient.setProperty("value", v)
        else:
            self.parameterWidget.qsb_expansion_coefficient.setProperty("rawValue", 0.0)
        if "ThermalExpansionReferenceTemperature" in matmap:
            self.parameterWidget.qsb_expansion_reference_temperature.setProperty(
                "value", Units.Quantity(matmap["ThermalExpansionReferenceTemperature"])
            )
        else:
            self.parameterWidget.qsb_expansion_reference_temperature.setProperty("rawValue", 0.0)
        if "SpecificHeat" in matmap:
            self.parameterWidget.qsb_specific_heat.setProperty(
                "value", Units.Quantity(matmap["SpecificHeat"])
            )
        else:
            self.parameterWidget.qsb_specific_heat.setProperty("rawValue", 0.0)

    def set_from_editor(self, value):
        if not value:
            return
        mat = self.material_manager.getMaterial(value)
        self.material = mat.Properties
        self.uuid = mat.UUID
        self.set_mat_params_in_input_fields(self.material)

    def mat_from_input_fields(self):
        d = {}
        d["Name"] = "Custom"
        p = self.parameterWidget
        d["Density"] = p.qsb_density.property("value").UserString
        d["ThermalConductivity"] = p.qsb_thermal_conductivity.property("value").UserString
        d["ThermalExpansionCoefficient"] = p.qsb_expansion_coefficient.property("value").UserString
        d["ThermalExpansionReferenceTemperature"] = p.qsb_expansion_reference_temperature.property(
            "value"
        ).UserString
        d["SpecificHeat"] = p.qsb_specific_heat.property("value").UserString
        if self.obj.Category == "Solid":
            d["YoungsModulus"] = p.qsb_young_modulus.property("value").UserString
            d["PoissonRatio"] = str(p.qsb_poisson_ratio.property("value").Value)
        elif self.obj.Category == "Fluid":
            d["KinematicViscosity"] = p.qsb_kinematic_viscosity.property("value").UserString
        self.material = d



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import uuid
import json
from typing import Mapping, List, Optional
import pathlib
import FreeCAD
import Path
from ...assets import Asset, AssetUri, AssetSerializer
from ...toolbit import ToolBit
from ..models.library import Library


class FCTLSerializer(AssetSerializer):
    for_class = Library
    extensions = (".fctl",)
    mime_type = "application/x-freecad-toolbit-library"

    @classmethod
    def get_label(cls) -> str:
        return FreeCAD.Qt.translate("CAM", "FreeCAD Tool Library")

    @classmethod
    def extract_dependencies(cls, data: bytes) -> List[AssetUri]:
        """Extracts URIs of dependencies from serialized data."""
        data_dict = json.loads(data.decode("utf-8"))
        tools_list = data_dict.get("tools", [])
        tool_ids = [pathlib.Path(tool["path"]).stem for tool in tools_list]
        return [AssetUri(f"toolbit://{tool_id}") for tool_id in tool_ids]

    @classmethod
    def serialize(cls, asset: Asset) -> bytes:
        """Serializes a Library object into bytes."""
        if not isinstance(asset, Library):
            raise TypeError(f"Expected Library instance, got {type(asset).__name__}")
        attrs = asset.to_dict()
        return json.dumps(attrs, sort_keys=True, indent=2).encode("utf-8")

    @classmethod
    def deserialize(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
    ) -> Library:
        """
        Creates a Library instance from serialized data and resolved
        dependencies.
        """

        data_dict = json.loads(data.decode("utf-8"))
        # The id parameter from the Asset.from_bytes method is the canonical ID
        # for the asset being deserialized. We should use this ID for the library
        # instance, overriding any 'id' that might be in the data_dict (which
        # is from an older version of the format).

        # For the label, prefer data_dict["label"], then "name", then fallback to using the id as filename
        # The id parameter often contains the filename stem when importing from files
        label = data_dict.get("label") or data_dict.get("name") or id or "Unnamed Library"
        library = Library(label, id=id)

        if dependencies is None:
            Path.Log.debug(
                f"FCTLSerializer.deserialize: Shallow load for library '{library.label}' (id: {id}). Tools not populated."
            )
            return library  # Only process tools if dependencies were resolved

        tools_list = data_dict.get("tools", [])
        for tool_data in tools_list:
            try:
                tool_no = int(tool_data["nr"])
            except ValueError:
                Path.Log.warning(f"Invalid tool ID in tool data: {tool_data}. Skipping.")
                continue
            tool_id = pathlib.Path(tool_data["path"]).stem  # Extract tool ID
            tool_uri = AssetUri(f"toolbit://{tool_id}")
            tool = dependencies.get(tool_uri)
            if tool:
                # Ensure the dependency is a ToolBit instance
                if not isinstance(tool, ToolBit):
                    Path.Log.warning(
                        f"Dependency for tool '{tool_id}' is not a ToolBit instance. Skipping."
                    )
                    continue
                library.add_bit(tool, bit_no=tool_no)
            else:
                # This should not happen if dependencies were resolved correctly,
                # but as a safeguard, log a warning and skip the tool.
                Path.Log.warning(
                    f"Tool with id {tool_id} not found in dependencies during deserialization."
                )
                # Create a placeholder toolbit with the original ID to preserve library structure
                from ...toolbit.models.custom import ToolBitCustom
                from ...shape.models.custom import ToolBitShapeCustom

                placeholder_shape = ToolBitShapeCustom(tool_id)
                placeholder_toolbit = ToolBitCustom(placeholder_shape, id=tool_id)
                placeholder_toolbit.label = f"Missing Tool ({tool_id})"
                library.add_bit(placeholder_toolbit, bit_no=tool_no)
                Path.Log.info(f"Created placeholder toolbit with original ID {tool_id}")
        return library

    @classmethod
    def deep_deserialize(cls, data: bytes) -> Library:
        """Deep deserialize a library by fetching all toolbit dependencies."""
        import uuid
        from ...camassets import cam_assets

        # Generate a unique ID for this library instance
        library_id = str(uuid.uuid4())

        Path.Log.info(
            f"FCTL DEEP_DESERIALIZE: Starting deep deserialization for library id='{library_id}'"
        )

        # Extract dependency URIs from the library data
        dependency_uris = cls.extract_dependencies(data)
        Path.Log.info(
            f"FCTL DEEP_DESERIALIZE: Found {len(dependency_uris)} toolbit dependencies: {[uri.asset_id for uri in dependency_uris]}"
        )

        # Fetch all toolbit dependencies
        resolved_dependencies = {}
        for dep_uri in dependency_uris:
            try:
                Path.Log.info(
                    f"FCTL DEEP_DESERIALIZE: Fetching toolbit '{dep_uri.asset_id}' from stores ['local', 'builtin']"
                )

                # Check if toolbit exists in each store individually for debugging
                exists_local = cam_assets.exists(dep_uri, store="local")
                exists_builtin = cam_assets.exists(dep_uri, store="builtin")
                Path.Log.info(
                    f"FCTL DEEP_DESERIALIZE: Toolbit '{dep_uri.asset_id}' exists - local: {exists_local}, builtin: {exists_builtin}"
                )

                toolbit = cam_assets.get(dep_uri, store=["local", "builtin"], depth=0)
                resolved_dependencies[dep_uri] = toolbit
                Path.Log.info(
                    f"FCTL DEEP_DESERIALIZE: Successfully fetched toolbit '{dep_uri.asset_id}'"
                )
            except Exception as e:
                Path.Log.warning(
                    f"FCTL DEEP_DESERIALIZE: Failed to fetch toolbit '{dep_uri.asset_id}': {e}"
                )

                # Try to get more detailed error information
                try:
                    # Check what's actually in the stores
                    local_toolbits = cam_assets.list_assets("toolbit", store="local")
                    local_ids = [uri.asset_id for uri in local_toolbits]
                    Path.Log.info(
                        f"FCTL DEBUG: Local store has {len(local_ids)} toolbits: {local_ids[:10]}{'...' if len(local_ids) > 10 else ''}"
                    )

                    if dep_uri.asset_id in local_ids:
                        Path.Log.warning(
                            f"FCTL DEBUG: Toolbit '{dep_uri.asset_id}' IS in local store list but get() failed!"
                        )
                except Exception as list_error:
                    Path.Log.error(f"FCTL DEBUG: Failed to list local toolbits: {list_error}")

        Path.Log.info(
            f"FCTL DEEP_DESERIALIZE: Resolved {len(resolved_dependencies)} of {len(dependency_uris)} dependencies"
        )

        # Now deserialize with the resolved dependencies
        return cls.deserialize(data, library_id, resolved_dependencies)

    @classmethod
    def deep_deserialize_with_context(cls, data: bytes, file_path: "pathlib.Path"):
        """Deep deserialize a library with file path context for external dependencies."""
        import uuid
        import pathlib
        from ...camassets import cam_assets
        from ...toolbit.serializers import all_serializers as toolbit_serializers

        # Use filename stem as library ID for meaningful names
        library_id = file_path.stem

        Path.Log.info(
            f"FCTL DEEP_DESERIALIZE_WITH_CONTEXT: Starting deep deserialization for library from {file_path}"
        )

        # Extract dependency URIs from the library data
        dependency_uris = cls.extract_dependencies(data)
        Path.Log.info(
            f"FCTL DEEP_DESERIALIZE_WITH_CONTEXT: Found {len(dependency_uris)} toolbit dependencies: {[uri.asset_id for uri in dependency_uris]}"
        )

        # Fetch all toolbit dependencies
        resolved_dependencies = {}
        for dep_uri in dependency_uris:
            try:
                # First try to get from asset manager stores
                Path.Log.info(
                    f"FCTL EXTERNAL: Trying to fetch toolbit '{dep_uri.asset_id}' from stores ['local', 'builtin']"
                )
                toolbit = cam_assets.get(dep_uri, store=["local", "builtin"], depth=0)
                resolved_dependencies[dep_uri] = toolbit
                Path.Log.info(
                    f"FCTL EXTERNAL: Successfully fetched toolbit '{dep_uri.asset_id}' from stores"
                )
            except Exception as e:
                # If not in stores, try to load from parallel Bit directory
                Path.Log.info(
                    f"FCTL EXTERNAL: Toolbit '{dep_uri.asset_id}' not in stores, trying external file: {e}"
                )

                # Look for toolbit files in parallel Bit directory
                library_dir = file_path.parent  # e.g., /path/to/Library/
                tools_dir = library_dir.parent  # e.g., /path/to/
                bit_dir = tools_dir / "Bit"  # e.g., /path/to/Bit/

                toolbit_loaded = False
                if bit_dir.exists():
                    possible_extensions = [".fctb", ".json", ".yaml", ".yml"]
                    for ext in possible_extensions:
                        toolbit_file = bit_dir / f"{dep_uri.asset_id}{ext}"
                        if toolbit_file.exists():
                            try:
                                # Find appropriate serializer for the file
                                from ...assets.ui.util import get_serializer_from_extension

                                serializer_class = get_serializer_from_extension(
                                    toolbit_serializers, ext, for_import=True
                                )
                                if serializer_class:
                                    # Load and deserialize the toolbit
                                    raw_toolbit_data = toolbit_file.read_bytes()
                                    toolbit = serializer_class.deep_deserialize(raw_toolbit_data)
                                    resolved_dependencies[dep_uri] = toolbit
                                    toolbit_loaded = True
                                    Path.Log.info(
                                        f"FCTL EXTERNAL: Successfully loaded toolbit '{dep_uri.asset_id}' from {toolbit_file}"
                                    )
                                    break
                            except Exception as load_error:
                                Path.Log.warning(
                                    f"FCTL EXTERNAL: Failed to load toolbit from {toolbit_file}: {load_error}"
                                )
                                continue

                if not toolbit_loaded:
                    Path.Log.warning(
                        f"FCTL EXTERNAL: Could not load toolbit '{dep_uri.asset_id}' from external files"
                    )

        Path.Log.info(
            f"FCTL EXTERNAL: Resolved {len(resolved_dependencies)} of {len(dependency_uris)} dependencies"
        )

        # Now deserialize with the resolved dependencies
        return cls.deserialize(data, library_id, resolved_dependencies)



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD

from . import manager
from .ccx_cantilever_base_edge import setup_cantilever_base_edge
from .manager import init_doc


def get_information():
    return {
        "name": "CCX cantilever beam rectangle",
        "meshtype": "edge",
        "meshelement": "Seg3",
        "constraints": ["fixed", "force"],
        "solvers": ["ccxtools"],
        "material": "solid",
        "equations": ["mechanical"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.ccx_cantilever_beam_rect import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=18&t=16044

CalculiX cantilever:
- modeled with seg3 beam elements

CrossSection:
- rectangle
- width = 400 mm
- height = 1250 mm

max deflection:
doc = App.ActiveDocument
len = doc.CantileverLine.Shape.Length
iyy = doc.CrossSectionFace.Shape.MatrixOfInertia.A22  # Iyy
force = doc.ConstraintForce.Force
from FreeCAD import Units
ym = Units.Quantity(doc.MechanicalMaterial.Material["YoungsModulus"]).getValueAs("MPa")
w = force * len**3 / (3 * ym * iyy)
w  # should print 112.3 mm

CalculiX FEM max deflection:
- 112.2 mm
- but the rotation seems 90 degree rotated (FIXME)

"""
    )


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    width = 400
    height = 1250
    cs_face = doc.addObject("Part::Plane", "CrossSectionFace")
    cs_face.Width = width
    cs_face.Length = height
    cs_face.Placement = FreeCAD.Placement(
        FreeCAD.Vector(0, 500 - 0.5 * width, 500 + 0.5 * height),
        FreeCAD.Rotation(0, 90, 0),
        FreeCAD.Vector(0, 0, 0),
    )
    doc.recompute()

    # setup CalculiX cantilever
    doc = setup_cantilever_base_edge(doc, solvertype)
    beamsection_obj = doc.getObject("BeamCrossSection")

    # change cross section to a circular section
    beamsection_obj.SectionType = "Rectangular"
    beamsection_obj.RectWidth = width
    beamsection_obj.RectHeight = height

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2023 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This module contains IFC-related FreeCAD commands"""

import FreeCAD
import FreeCADGui

from . import ifc_openshell

translate = FreeCAD.Qt.translate
QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP


def get_project():
    """Gets the current project"""

    from . import ifc_tools

    if FreeCADGui.Selection.getSelection():
        return ifc_tools.get_project(FreeCADGui.Selection.getSelection()[0])
    else:
        return ifc_tools.get_project(FreeCAD.ActiveDocument)


class IFC_Diff:
    """Shows a diff of the changes in the current IFC document"""

    def GetResources(self):
        tt = QT_TRANSLATE_NOOP("IFC_Diff", "Shows the current unsaved changes in the IFC file")
        return {
            "Pixmap": "IFC",
            "MenuText": QT_TRANSLATE_NOOP("IFC_Diff", "IFC Diff"),
            "ToolTip": tt,
            "Accel": "I, D",
        }

    def Activated(self):
        from . import ifc_diff

        proj = get_project()
        if proj:
            diff = ifc_diff.get_diff(proj)
            ifc_diff.show_diff(diff)


class IFC_Expand:
    """Expands the children of the selected objects or document"""

    def GetResources(self):
        tt = QT_TRANSLATE_NOOP(
            "IFC_Expand", "Expands the children of the selected objects or document"
        )
        return {
            "Pixmap": "IFC",
            "MenuText": QT_TRANSLATE_NOOP("IFC_Expand", "IFC Expand"),
            "ToolTip": tt,
            "Accel": "I, E",
        }

    def Activated(self):
        ns = []
        for obj in FreeCADGui.Selection.getSelection():
            if hasattr(obj.ViewObject, "Proxy"):
                if hasattr(obj.ViewObject.Proxy, "hasChildren"):
                    if obj.ViewObject.Proxy.hasChildren(obj):
                        no = obj.ViewObject.Proxy.expandChildren(obj)
                        ns.extend(no)
        else:
            from . import ifc_generator
            from . import ifc_tools

            document = FreeCAD.ActiveDocument
            ifc_generator.delete_ghost(document)
            ifcfile = ifc_tools.get_ifcfile(document)
            if ifcfile:
                ns = ifc_tools.create_children(
                    document, ifcfile, recursive=True, only_structure=True
                )
        if ns:
            document.recompute()
            FreeCADGui.Selection.clearSelection()
            for o in ns:
                FreeCADGui.Selection.addSelection(o)


class IFC_ConvertDocument:
    """Converts the active document to an IFC document"""

    def GetResources(self):
        tt = QT_TRANSLATE_NOOP(
            "IFC_ConvertDocument", "Converts the active document to an IFC document"
        )
        return {
            "Pixmap": "IFC",
            "MenuText": QT_TRANSLATE_NOOP("IFC_ConvertDocument", "Convert Document"),
            "ToolTip": tt,
            # "Accel": "I, C",
        }

    def Activated(self):
        doc = FreeCAD.ActiveDocument
        if hasattr(doc, "Proxy") and hasattr(doc.Proxy, "ifcfile") and doc.Proxy.ifcfile:
            FreeCAD.Console.PrintError(
                translate("BIM", "The active document is already an IFC document")
            )
        else:
            from . import ifc_tools

            ifc_tools.convert_document(doc)


class IFC_MakeProject:
    """Converts the current selection to an IFC project"""

    def GetResources(self):
        tt = QT_TRANSLATE_NOOP(
            "IFC_MakeProject", "Converts the current selection to an IFC project"
        )
        return {
            "Pixmap": "IFC",
            "MenuText": QT_TRANSLATE_NOOP("IFC_MakeProject", "Convert to IFC Project"),
            "ToolTip": tt,
            "Accel": "I, P",
        }

    def IsActive(self):
        return bool(FreeCADGui.Selection.getSelection())

    def Activated(self):
        from importers import exportIFC  # lazy loading
        from . import ifc_tools
        from PySide import QtGui

        doc = FreeCAD.ActiveDocument
        objs = FreeCADGui.Selection.getSelection()
        sf = QtGui.QFileDialog.getSaveFileName(
            None,
            "Save an IFC file",
            None,
            "Industry Foundation Classes (*.ifc)",
        )
        if sf and sf[0]:
            sf = sf[0]
            if not sf.lower().endswith(".ifc"):
                sf += ".ifc"
            exportIFC.export(objs, sf)
            ifc_tools.create_document_object(doc, sf, strategy=2)
            ifc_tools.remove_tree(objs)
            doc.recompute()


class IFC_Save:
    """Saves the current IFC document"""

    def GetResources(self):
        tt = QT_TRANSLATE_NOOP("IFC_Save", "Saves the current IFC document")
        return {
            "Pixmap": "IFC_document",
            "MenuText": QT_TRANSLATE_NOOP("IFC_Save", "Save IFC File"),
            "ToolTip": tt,
            "Accel": "Ctrl+S",
        }

    def IsActive(self):
        doc = FreeCAD.ActiveDocument
        if hasattr(doc, "IfcFilePath"):
            return True
        return False

    def Activated(self):
        from . import ifc_tools  # lazy loading

        doc = FreeCAD.ActiveDocument
        if getattr(doc, "IfcFilePath", None):
            ifc_tools.save(doc)
            gdoc = FreeCADGui.getDocument(doc.Name)
            try:
                gdoc.Modified = False
            except:
                pass
        else:
            FreeCADGui.runCommand("IFC_SaveAs")


class IFC_SaveAs:
    """Saves the current IFC document as another name"""

    def GetResources(self):
        tt = QT_TRANSLATE_NOOP("IFC_SaveAs", "Saves the current IFC document as another file")
        return {
            "Pixmap": "IFC_document",
            "MenuText": QT_TRANSLATE_NOOP("IFC_SaveAs", "Save IFC File As…"),
            "ToolTip": tt,
            "Accel": "Ctrl+Shift+S",
        }

    def IsActive(self):
        doc = FreeCAD.ActiveDocument
        if hasattr(doc, "IfcFilePath"):
            return True
        return False

    def Activated(self):
        from . import ifc_tools  # lazy loading
        from . import ifc_viewproviders

        doc = FreeCAD.ActiveDocument
        if ifc_viewproviders.get_filepath(doc):
            ifc_tools.save(doc)
            gdoc = FreeCADGui.getDocument(doc.Name)
            try:
                gdoc.Modified = False
            except:
                pass


def get_commands():
    """Returns a list of IFC commands"""

    return ["IFC_Diff", "IFC_Expand", "IFC_MakeProject", "IFC_UpdateIOS"]


# initialize commands
FreeCADGui.addCommand("IFC_Diff", IFC_Diff())
FreeCADGui.addCommand("IFC_Expand", IFC_Expand())
FreeCADGui.addCommand("IFC_ConvertDocument", IFC_ConvertDocument())
FreeCADGui.addCommand("IFC_MakeProject", IFC_MakeProject())
FreeCADGui.addCommand("IFC_Save", IFC_Save())
FreeCADGui.addCommand("IFC_SaveAs", IFC_SaveAs())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2023 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This NativeIFC module deals with materials"""

import ifcopenshell
import ifcopenshell.util.element

import FreeCAD

from . import ifc_tools


def create_material(element, parent, recursive=False):
    """Creates a material object in the given project or parent material"""

    if not element:
        return
    if isinstance(element, (tuple, list)):
        for e in element:
            create_material(e, parent, recursive)
        return
    if hasattr(parent, "Document"):
        doc = parent.Document
    else:
        doc = parent
    exobj = ifc_tools.get_object(element, doc)
    if exobj:
        return exobj
    obj = ifc_tools.add_object(doc, otype="material")
    ifcfile = ifc_tools.get_ifcfile(parent)
    ifc_tools.add_properties(obj, ifcfile, element)
    if parent.isDerivedFrom("App::MaterialObject"):
        parent.Proxy.addObject(parent, obj)
    else:
        ifc_tools.get_group(parent, "IfcMaterialsGroup").addObject(obj)
    if recursive:
        submat = get_material(obj)
        if isinstance(submat, list):
            for s in submat:
                create_material(s, obj, recursive)
        else:
            create_material(submat, obj, recursive)
    return obj


def show_material(obj):
    """Creates and links materials for the given object, if available"""

    material = get_material(obj)
    if not material:
        return
    if not hasattr(obj, "Material"):
        obj.addProperty("App::PropertyLink", "Material", "IFC", locked=True)
    project = ifc_tools.get_project(obj)
    matobj = create_material(material, project, recursive=True)
    obj.Material = matobj


def load_materials(obj):
    """Recursively loads materials of child objects"""

    show_material(obj)
    if isinstance(obj, FreeCAD.DocumentObject):
        group = obj.Group
    else:
        group = obj.Objects
    for child in group:
        load_materials(child)


def get_material(obj):
    """Returns a material attached to this object"""

    element = ifc_tools.get_ifc_element(obj)
    if not element:
        return None
    if element.is_a("IfcMaterialConstituentSet"):
        return element.MaterialConstituents
    elif element.is_a() in [
        "IfcMaterialLayer",
        "IfcMaterialConstituent",
        "IfcMaterialProfile",
    ]:
        return element.Material
    elif element.is_a("IfcMaterialLayerSet"):
        return element.MaterialLayers
    elif element.is_a("IfcMaterialProfileSet"):
        return element.MaterialProfiles
    else:
        material = ifcopenshell.util.element.get_material(element, should_skip_usage=True)
        return material


def set_material(material, obj):
    """Attributes a material to an object"""

    ifcfile = ifc_tools.get_ifcfile(obj)
    element = ifc_tools.get_ifc_element(obj)
    material_element = ifc_tools.get_ifc_element(material)
    if not ifcfile:
        return
    new = False
    if not material_element or ifc_tools.get_ifcfile(material) != ifcfile:
        material_element = ifc_tools.api_run("material.add_material", ifcfile, name=material.Label)
        new = True
        delete = not (ifc_tools.PARAMS.GetBool("KeepAggregated", False))
        if delete and len(material.InList) == 1:
            container = material.InList[0]
            doc = material.Document
            doc.removeObject(material.Name)
            if not container.OutList:
                doc.removeObject(container.Name)
    if material_element:
        try:
            # IfcOpenShell 0.8
            ifc_tools.api_run(
                "material.assign_material",
                ifcfile,
                products=[element],
                type=material_element.is_a(),
                material=material_element,
            )
        except:
            # IfcOpenShell 0.7
            ifc_tools.api_run(
                "material.assign_material",
                ifcfile,
                product=element,
                type=material_element.is_a(),
                material=material_element,
            )
        if new:
            show_material(obj)



# ***************************************************************************
# *   Copyright (c) 2016 Qingfeng Xia <qingfeng.xia()eng.ox.ac.uk>          *
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM result mechanical document object"
__author__ = "Qingfeng Xia, Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package result_mechanical
#  \ingroup FEM
#  \brief mechanical result object

from . import base_fempythonobject


class ResultMechanical(base_fempythonobject.BaseFemPythonObject):
    """
    The Fem::ResultMechanical's Proxy python type, add result specific properties
    """

    Type = "Fem::ResultMechanical"

    def __init__(self, obj):
        super().__init__(obj)

        obj.addProperty(
            "App::PropertyString",
            "ResultType",
            "Base",
            "Type of the result",
            1,  # the 1 set the property to ReadOnly
        )
        obj.setPropertyStatus("ResultType", "LockDynamic")
        obj.ResultType = str(self.Type)

        # for frequency analysis
        obj.addProperty("App::PropertyInteger", "Eigenmode", "Data", "", True)
        obj.setPropertyStatus("Eigenmode", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloat", "EigenmodeFrequency", "Data", "User Defined Results", True
        )
        obj.setPropertyStatus("EigenmodeFrequency", "LockDynamic")

        # node results
        # set read only or hide a property:
        # https://forum.freecad.org/viewtopic.php?f=18&t=13460&start=10#p108072
        # do not show up in propertyEditor of comboView
        obj.addProperty(
            "App::PropertyVectorList",
            "DisplacementVectors",
            "NodeData",
            "List of displacement vectors",
            True,
        )
        obj.setPropertyStatus("DisplacementVectors", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloatList",
            "Peeq",
            "NodeData",
            "List of equivalent plastic strain values",
            True,
        )
        obj.setPropertyStatus("Peeq", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloatList",
            "MohrCoulomb",
            "NodeData",
            "List of Mohr Coulomb stress values",
            True,
        )
        obj.setPropertyStatus("MohrCoulomb", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloatList",
            "ReinforcementRatio_x",
            "NodeData",
            "Reinforcement ratio in X-direction",
            True,
        )
        obj.setPropertyStatus("ReinforcementRatio_x", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloatList",
            "ReinforcementRatio_y",
            "NodeData",
            "Reinforcement ratio in Y-direction",
            True,
        )
        obj.setPropertyStatus("ReinforcementRatio_y", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloatList",
            "ReinforcementRatio_z",
            "NodeData",
            "Reinforcement ratio in Z-direction",
            True,
        )
        obj.setPropertyStatus("ReinforcementRatio_z", "LockDynamic")
        # these three principal vectors are used only if there is a reinforced mat obj
        # https://forum.freecad.org/viewtopic.php?f=18&t=33106&p=416006#p416006
        obj.addProperty(
            "App::PropertyVectorList",
            "PS1Vector",
            "NodeData",
            "List of 1st Principal Stress Vectors",
            True,
        )
        obj.setPropertyStatus("PS1Vector", "LockDynamic")
        obj.addProperty(
            "App::PropertyVectorList",
            "PS2Vector",
            "NodeData",
            "List of 2nd Principal Stress Vectors",
            True,
        )
        obj.setPropertyStatus("PS2Vector", "LockDynamic")
        obj.addProperty(
            "App::PropertyVectorList",
            "PS3Vector",
            "NodeData",
            "List of 3rd Principal Stress Vectors",
            True,
        )
        obj.setPropertyStatus("PS3Vector", "LockDynamic")

        # readonly in propertyEditor of comboView
        obj.addProperty(
            "App::PropertyFloatList",
            "DisplacementLengths",
            "NodeData",
            "List of displacement lengths",
            True,
        )
        obj.setPropertyStatus("DisplacementLengths", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloatList",
            "vonMises",
            "NodeData",
            "List of von Mises equivalent stresses",
            True,
        )
        obj.setPropertyStatus("vonMises", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "PrincipalMax", "NodeData", "", True)
        obj.setPropertyStatus("PrincipalMax", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "PrincipalMed", "NodeData", "", True)
        obj.setPropertyStatus("PrincipalMed", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "PrincipalMin", "NodeData", "", True)
        obj.setPropertyStatus("PrincipalMin", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloatList",
            "MaxShear",
            "NodeData",
            "List of Maximum Shear stress values",
            True,
        )
        obj.setPropertyStatus("MaxShear", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloatList",
            "MassFlowRate",
            "NodeData",
            "List of mass flow rate values",
            True,
        )
        obj.setPropertyStatus("MassFlowRate", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloatList",
            "NetworkPressure",
            "NodeData",
            "List of network pressure values",
            True,
        )
        obj.setPropertyStatus("NetworkPressure", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloatList", "UserDefined", "NodeData", "User Defined Results", True
        )
        obj.setPropertyStatus("UserDefined", "LockDynamic")
        obj.addProperty(
            "App::PropertyFloatList", "Temperature", "NodeData", "Temperature field", True
        )
        obj.addProperty(
            "App::PropertyVectorList", "HeatFlux", "NodeData", "List of heat flux vectors", True
        )
        obj.setPropertyStatus("HeatFlux", "LockDynamic")

        obj.setPropertyStatus("Temperature", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStressXX", "NodeData", "", True)
        obj.setPropertyStatus("NodeStressXX", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStressYY", "NodeData", "", True)
        obj.setPropertyStatus("NodeStressYY", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStressZZ", "NodeData", "", True)
        obj.setPropertyStatus("NodeStressZZ", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStressXY", "NodeData", "", True)
        obj.setPropertyStatus("NodeStressXY", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStressXZ", "NodeData", "", True)
        obj.setPropertyStatus("NodeStressXZ", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStressYZ", "NodeData", "", True)
        obj.setPropertyStatus("NodeStressYZ", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStrainXX", "NodeData", "", True)
        obj.setPropertyStatus("NodeStrainXX", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStrainYY", "NodeData", "", True)
        obj.setPropertyStatus("NodeStrainYY", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStrainZZ", "NodeData", "", True)
        obj.setPropertyStatus("NodeStrainZZ", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStrainXY", "NodeData", "", True)
        obj.setPropertyStatus("NodeStrainXY", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStrainXZ", "NodeData", "", True)
        obj.setPropertyStatus("NodeStrainXZ", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "NodeStrainYZ", "NodeData", "", True)
        obj.setPropertyStatus("NodeStrainYZ", "LockDynamic")
        obj.addProperty("App::PropertyFloatList", "CriticalStrainRatio", "NodeData", "", True)
        obj.setPropertyStatus("CriticalStrainRatio", "LockDynamic")

        # initialize the Stats with the appropriate count of items
        # see fill_femresult_stats in femresult/resulttools.py
        zero_list = 26 * [0]
        obj.Stats = zero_list

    def onDocumentRestored(self, obj):
        # migrate old result objects, because property "StressValues"
        # was renamed to "vonMises" in commit 8b68ab7
        if hasattr(obj, "StressValues") is True:
            obj.addProperty(
                "App::PropertyFloatList",
                "vonMises",
                "NodeData",
                "List of von Mises equivalent stresses",
                True,
            )
            obj.setPropertyStatus("vonMises", "LockDynamic")
            obj.vonMises = obj.StressValues
            obj.setPropertyStatus("StressValues", "-LockDynamic")
            obj.removeProperty("StressValues")

        # migrate old result objects, because property "Stats"
        # consisting of min, avg, max values was reduced to min, max in commit c2a57b3e
        if len(obj.Stats) == 39:
            temp = obj.Stats
            for i in range(12, -1, -1):
                del temp[3 * i + 1]
            obj.Stats = temp



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Defines a QWidget-derived class for displaying the view selection buttons."""

from enum import IntEnum
import os
import addonmanager_freecad_interface as fci
from PySideWrapper import QtCore, QtGui, QtWidgets

translate = fci.translate


class AddonManagerDisplayStyle(IntEnum):
    """The display mode of the Addon Manager"""

    COMPACT = 0
    EXPANDED = 1
    COMPOSITE = 2


class WidgetViewSelector(QtWidgets.QWidget):
    """A widget for selecting the Addon Manager's primary view mode"""

    view_changed = QtCore.Signal(int)

    def __init__(self, parent: QtWidgets.QWidget = None):
        super().__init__(parent)
        self.horizontal_layout = None
        self.composite_button = None
        self.expanded_button = None
        self.compact_button = None
        self._setup_ui()
        self._setup_connections()

    def set_current_view(self, view: AddonManagerDisplayStyle):
        """Set the current selection. Does NOT emit a view_changed signal, only changes the
        interface display."""
        self.compact_button.setChecked(False)
        self.expanded_button.setChecked(False)
        self.composite_button.setChecked(False)
        if view == AddonManagerDisplayStyle.COMPACT:
            self.compact_button.setChecked(True)
        elif view == AddonManagerDisplayStyle.EXPANDED:
            self.expanded_button.setChecked(True)
        elif view == AddonManagerDisplayStyle.COMPOSITE:
            self.composite_button.setChecked(True)
        else:
            fci.Console.PrintWarning(f"Unrecognized display style {view}")

    def _setup_ui(self):
        icon_path = os.path.join(os.path.dirname(__file__), "..", "Resources", "icons")
        self.horizontal_layout = QtWidgets.QHBoxLayout()
        self.horizontal_layout.setContentsMargins(0, 0, 0, 0)
        self.horizontal_layout.setSpacing(2)
        self.compact_button = QtWidgets.QToolButton(self)
        self.compact_button.setObjectName("compact_button")
        self.compact_button.setIcon(
            QtGui.QIcon.fromTheme("back", QtGui.QIcon(os.path.join(icon_path, "compact_view.svg")))
        )
        self.compact_button.setCheckable(True)
        self.compact_button.setAutoExclusive(True)

        self.expanded_button = QtWidgets.QToolButton(self)
        self.expanded_button.setObjectName("expanded_button")
        self.expanded_button.setCheckable(True)
        self.expanded_button.setChecked(True)
        self.expanded_button.setAutoExclusive(True)
        self.expanded_button.setIcon(
            QtGui.QIcon.fromTheme(
                "expanded_view", QtGui.QIcon(os.path.join(icon_path, "expanded_view.svg"))
            )
        )

        self.composite_button = QtWidgets.QToolButton(self)
        self.composite_button.setObjectName("composite_button")
        if (
            QtCore.QLibraryInfo.version().majorVersion() == 5
            and QtCore.QLibraryInfo.version().minorVersion() < 15
        ):
            self.composite_button.setEnabled(False)
            self.composite_button.setCheckable(False)
            self.composite_button.setChecked(False)
        else:
            self.composite_button.setCheckable(True)
            self.composite_button.setChecked(True)
        self.composite_button.setAutoExclusive(True)
        self.composite_button.setIcon(
            QtGui.QIcon.fromTheme(
                "composite_button", QtGui.QIcon(os.path.join(icon_path, "composite_view.svg"))
            )
        )
        self.horizontal_layout.addWidget(self.compact_button)
        self.horizontal_layout.addWidget(self.expanded_button)
        self.horizontal_layout.addWidget(self.composite_button)

        self.compact_button.clicked.connect(
            lambda: self.view_changed.emit(AddonManagerDisplayStyle.COMPACT)
        )
        self.expanded_button.clicked.connect(
            lambda: self.view_changed.emit(AddonManagerDisplayStyle.EXPANDED)
        )
        self.composite_button.clicked.connect(
            lambda: self.view_changed.emit(AddonManagerDisplayStyle.COMPOSITE)
        )

        self.setLayout(self.horizontal_layout)
        self.retranslateUi(None)

    def _setup_connections(self):
        self.compact_button.clicked.connect(
            lambda: self.view_changed.emit(AddonManagerDisplayStyle.COMPACT)
        )
        self.expanded_button.clicked.connect(
            lambda: self.view_changed.emit(AddonManagerDisplayStyle.EXPANDED)
        )
        self.composite_button.clicked.connect(
            lambda: self.view_changed.emit(AddonManagerDisplayStyle.COMPOSITE)
        )

    def retranslateUi(self, _):
        self.composite_button.setToolTip(translate("AddonsInstaller", "Composite view"))
        self.expanded_button.setToolTip(translate("AddonsInstaller", "Expanded view"))
        self.compact_button.setToolTip(translate("AddonsInstaller", "Compact view"))



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2023 Werner Mayer <wmayer[at]users.sourceforge.net>     *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "BOPTools.BOPFeatures module"
__author__ = "Werner Mayer"
__url__ = "https://www.freecad.org"
__doc__ = "Helper class to create the features for Boolean operations."

import FreeCAD
import Part


class BOPFeatures:
    def __init__(self, doc):
        self.doc = doc

    def make_section(self, inputNames):
        obj = self.doc.addObject("Part::Section", "Section")
        obj.Base = self.doc.getObject(inputNames[0])
        obj.Tool = self.doc.getObject(inputNames[1])
        self.copy_visual_attributes(obj, obj.Base)
        target = self.move_input_objects([obj.Base, obj.Tool])
        if target:
            target.addObject(obj)
        return obj

    def make_cut(self, inputNames):
        obj = self.doc.addObject("Part::Cut", "Cut")
        obj.Base = self.doc.getObject(inputNames[0])
        obj.Tool = self.doc.getObject(inputNames[1])
        self.copy_visual_attributes(obj, obj.Base)
        target = self.move_input_objects([obj.Base, obj.Tool])
        if target:
            target.addObject(obj)
        return obj

    def make_common(self, inputNames):
        obj = self.doc.addObject("Part::Common", "Common")
        obj.Base = self.doc.getObject(inputNames[0])
        obj.Tool = self.doc.getObject(inputNames[1])
        self.copy_visual_attributes(obj, obj.Base)
        target = self.move_input_objects([obj.Base, obj.Tool])
        if target:
            target.addObject(obj)
        return obj

    def make_multi_common(self, inputNames):
        obj = self.doc.addObject("Part::MultiCommon", "Common")
        obj.Shapes = [self.doc.getObject(name) for name in inputNames]
        self.copy_visual_attributes(obj, obj.Shapes[0])
        target = self.move_input_objects(obj.Shapes)
        if target:
            target.addObject(obj)
        return obj

    def make_fuse(self, inputNames):
        obj = self.doc.addObject("Part::Fuse", "Fusion")
        obj.Base = self.doc.getObject(inputNames[0])
        obj.Tool = self.doc.getObject(inputNames[1])
        self.copy_visual_attributes(obj, obj.Base)
        target = self.move_input_objects([obj.Base, obj.Tool])
        if target:
            target.addObject(obj)
        return obj

    def make_multi_fuse(self, inputNames):
        obj = self.doc.addObject("Part::MultiFuse", "Fusion")
        obj.Shapes = [self.doc.getObject(name) for name in inputNames]
        self.copy_visual_attributes(obj, obj.Shapes[0])
        target = self.move_input_objects(obj.Shapes)
        if target:
            target.addObject(obj)
        return obj

    def move_input_objects(self, objects):
        targetGroup = None
        for obj in objects:
            obj.Visibility = False
            parent = obj.getParent()
            if parent:
                parent.removeObject(obj)
                targetGroup = parent
        return targetGroup

    def copy_visual_attributes(self, target, source):
        if target.ViewObject:
            displayMode = source.ViewObject.DisplayMode
            src = source
            while displayMode == "Link":
                if getattr(src, "LinkedObject", None):
                    src = src.LinkedObject
                elif getattr(src, "Base", None):
                    # Draft Link array
                    src = src.Base
                else:
                    break
                if not hasattr(src, "ViewObject"):
                    break
                displayMode = src.ViewObject.DisplayMode
            if displayMode in target.ViewObject.getEnumerationsOfProperty("DisplayMode"):
                target.ViewObject.DisplayMode = displayMode



# ***************************************************************************
# *   Copyright (c) 2017 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver Z88 writer"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import time
from os.path import join

import FreeCAD

from .. import writerbase
from feminout import importZ88Mesh
from femmesh import meshtools


class FemInputWriterZ88(writerbase.FemInputWriter):
    def __init__(self, analysis_obj, solver_obj, mesh_obj, member, dir_name=None):
        writerbase.FemInputWriter.__init__(
            self, analysis_obj, solver_obj, mesh_obj, member, dir_name
        )
        self.file_name = join(self.dir_name, "z88")

    # ********************************************************************************************
    # write solver input
    def write_solver_input(self):
        timestart = time.process_time()
        FreeCAD.Console.PrintMessage("\n")  # because of time print in separate line
        FreeCAD.Console.PrintMessage("Z88 solver input writing...\n")
        FreeCAD.Console.PrintLog(f"FemInputWriterZ88 --> self.dir_name  -->  {self.dir_name}\n")
        FreeCAD.Console.PrintMessage(
            f"FemInputWriterZ88 --> self.file_name  -->  {self.file_name}\n"
        )
        FreeCAD.Console.PrintMessage(f"Write z88 input files to: {self.dir_name}\n")
        control = self.set_z88_elparam()
        if control is False:
            return None
        self.write_z88_mesh()
        self.write_z88_constraints()
        self.write_z88_face_loads()
        self.write_z88_materials()
        self.write_z88_elements_properties()
        self.write_z88_integration_properties()
        self.write_z88_memory_parameter()
        self.write_z88_solver_parameter()
        writing_time_string = "Writing time input file: {} seconds".format(
            round((time.process_time() - timestart), 2)
        )
        FreeCAD.Console.PrintMessage(f"{writing_time_string}\n\n")
        return self.dir_name

    # ********************************************************************************************
    def set_z88_elparam(self):
        # TODO: param should be moved to the solver object like the known analysis
        z8804 = {"INTORD": "0", "INTOS": "0", "IHFLAG": "0", "ISFLAG": "1"}  # seg2 --> stab4
        z8824 = {"INTORD": "7", "INTOS": "7", "IHFLAG": "1", "ISFLAG": "1"}  # tria6 --> schale24
        z8823 = {"INTORD": "3", "INTOS": "0", "IHFLAG": "1", "ISFLAG": "0"}  # quad8 --> schale23
        z8817 = {"INTORD": "4", "INTOS": "0", "IHFLAG": "0", "ISFLAG": "0"}  # tetra4 --> volume17
        z8816 = {"INTORD": "4", "INTOS": "0", "IHFLAG": "0", "ISFLAG": "0"}  # tetra10 --> volume16
        z8801 = {"INTORD": "2", "INTOS": "2", "IHFLAG": "0", "ISFLAG": "1"}  # hexa8 --> volume1
        z8810 = {"INTORD": "3", "INTOS": "0", "IHFLAG": "0", "ISFLAG": "0"}  # hexa20 --> volume10
        param = {4: z8804, 24: z8824, 23: z8823, 17: z8817, 16: z8816, 1: z8801, 10: z8810}
        # TODO: test elements 17, 16, 10, INTORD etc
        self.z88_element_type = importZ88Mesh.get_z88_element_type(
            self.femmesh, self.femelement_table
        )
        if self.z88_element_type in param:
            self.z88_elparam = param[self.z88_element_type]
        else:
            FreeCAD.Console.PrintError(
                "Element type not supported by Z88. Can not write Z88 solver input.\n"
            )
            return False
        FreeCAD.Console.PrintMessage(self.z88_elparam)
        FreeCAD.Console.PrintMessage("\n")
        return True

    # ********************************************************************************************
    def write_z88_mesh(self):
        if not self.femnodes_mesh:
            self.femnodes_mesh = self.femmesh.Nodes
        if not self.femelement_table:
            self.femelement_table = meshtools.get_femelement_table(self.femmesh)
            self.element_count = len(self.femelement_table)
        mesh_file_path = self.file_name + "i1.txt"
        f = open(mesh_file_path, "w")
        importZ88Mesh.write_z88_mesh_to_file(
            self.femnodes_mesh, self.femelement_table, self.z88_element_type, f
        )
        f.close()

    # ********************************************************************************************
    def write_z88_constraints(self):
        constraints_data = []  # will be a list of tuple for better sorting

        # fixed constraints
        # write nodes to constraints_data (different from writing to file in ccxInpWriter
        for femobj in self.member.cons_fixed:
            for n in femobj["Nodes"]:
                constraints_data.append((n, f"{n}  1  2  0\n"))
                constraints_data.append((n, f"{n}  2  2  0\n"))
                constraints_data.append((n, f"{n}  3  2  0\n"))

        # forces constraints
        # write node loads to constraints_data
        # a bit different from writing to file for ccxInpWriter
        for femobj in self.member.cons_force:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            direction_vec = femobj["Object"].DirectionVector
            for ref_shape in femobj["NodeLoadTable"]:
                for n in sorted(ref_shape[1]):
                    # the loads in ref_shape[1][n] are without unit
                    node_load = ref_shape[1][n]
                    if direction_vec.x != 0.0:
                        v1 = direction_vec.x * node_load
                        constraints_data.append((n, f"{n}  1  1  {v1}\n"))
                    if direction_vec.y != 0.0:
                        v2 = direction_vec.y * node_load
                        constraints_data.append((n, f"{n}  2  1  {v2}\n"))
                    if direction_vec.z != 0.0:
                        v3 = direction_vec.z * node_load
                        constraints_data.append((n, f"{n}  3  1  {v3}\n"))

        # write constraints_data to file
        constraints_file_path = self.file_name + "i2.txt"
        f = open(constraints_file_path, "w")
        f.write(str(len(constraints_data)) + "\n")
        for c in sorted(constraints_data):
            f.write(c[1])
        f.close()

    # ********************************************************************************************
    def write_z88_face_loads(self):
        # not yet supported
        face_load_file_path = self.file_name + "i5.txt"
        f = open(face_load_file_path, "w")
        f.write(" 0")
        f.write("\n")
        f.close()

    # ********************************************************************************************
    def write_z88_materials(self):
        mat_obj = self.member.mats_linear[0]["Object"]
        material_data_file_name = "51.txt"
        materials_file_path = self.file_name + "mat.txt"
        fms = open(materials_file_path, "w")
        fms.write("1\n")
        fms.write(f"1 {self.element_count} {material_data_file_name}")
        fms.write("\n")
        fms.close()
        material_data_file_path = join(self.dir_name, material_data_file_name)
        fmd = open(material_data_file_path, "w")
        YM = FreeCAD.Units.Quantity(mat_obj.Material["YoungsModulus"])
        YM_in_MPa = YM.getValueAs("MPa")
        PR = float(mat_obj.Material["PoissonRatio"])
        fmd.write(f"{YM_in_MPa} {PR:.3f}")
        fmd.write("\n")
        fmd.close()

    # ********************************************************************************************
    def write_z88_elements_properties(self):
        element_properties_file_path = self.file_name + "elp.txt"
        elements_data = []
        if meshtools.is_edge_femmesh(self.femmesh):
            beam_obj = self.member.geos_beamsection[0]["Object"]
            area = 0
            if beam_obj.SectionType == "Rectangular":
                width = beam_obj.RectWidth.getValueAs("mm").Value
                height = beam_obj.RectHeight.getValueAs("mm").Value
                area = width * height
            elif beam_obj.SectionType == "Circular":
                diameter = beam_obj.CircDiameter.getValueAs("mm").Value
                from math import pi

                area = 0.25 * pi * diameter * diameter
            else:
                FreeCAD.Console.PrintError(
                    "Cross section type {} not supported, "
                    "cross section area will be 0 in solver input.\n".format(beam_obj.SectionType)
                )
                # TODO make the check in prechecks and delete it here
                # no extensive errorhandling in writer
                # this way the solver will fail and an exception is raised somehow
            elements_data.append(f"1 {self.element_count} {area} 0 0 0 0 0 0 ")
            FreeCAD.Console.PrintWarning("Be aware, only trusses are supported for edge meshes!\n")
        elif meshtools.is_face_femmesh(self.femmesh):
            thick_obj = self.member.geos_shellthickness[0]["Object"]
            thickness = thick_obj.Thickness.getValueAs("mm").Value
            elements_data.append(f"1 {self.element_count} {thickness} 0 0 0 0 0 0 ")
        elif meshtools.is_solid_femmesh(self.femmesh):
            elements_data.append(f"1 {self.element_count} 0 0 0 0 0 0 0")
        else:
            FreeCAD.Console.PrintError("Error!\n")
        f = open(element_properties_file_path, "w")
        f.write(str(len(elements_data)) + "\n")
        for e in elements_data:
            f.write(e)
        f.write("\n")
        f.close()

    # ********************************************************************************************
    def write_z88_integration_properties(self):
        integration_data = []
        integration_data.append(
            "1 {} {} {}".format(
                self.element_count, self.z88_elparam["INTORD"], self.z88_elparam["INTOS"]
            )
        )
        integration_properties_file_path = self.file_name + "int.txt"
        f = open(integration_properties_file_path, "w")
        f.write(f"{len(integration_data)}\n")
        for i in integration_data:
            f.write(i)
        f.write("\n")
        f.close()

    # ********************************************************************************************
    def write_z88_solver_parameter(self):
        global z88_man_template
        z88_man_template = z88_man_template.replace(
            "$z88_param_ihflag", str(self.z88_elparam["IHFLAG"])
        )
        z88_man_template = z88_man_template.replace(
            "$z88_param_isflag", str(self.z88_elparam["ISFLAG"])
        )
        solver_parameter_file_path = self.file_name + "man.txt"
        f = open(solver_parameter_file_path, "w")
        f.write(z88_man_template)
        f.close()

    # ********************************************************************************************
    def write_z88_memory_parameter(self):
        prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/Z88")
        MaxGS = prefs.GetInt("MaxGS", 100000000)
        MaxKOI = prefs.GetInt("MaxKOI", 2800000)
        global z88_dyn_template
        template_array = z88_dyn_template.splitlines()
        output = ""
        for line in template_array:
            if line.find("MAXGS") > -1:
                line = f"    MAXGS  {MaxGS}"
            if line.find("MAXKOI") > -1:
                line = f"    MAXKOI   {MaxKOI}"
            output += line + "\n"

        solver_parameter_file_path = self.file_name + ".dyn"
        f = open(solver_parameter_file_path, "w")
        f.write(output)
        f.close()


# for solver parameter file Z88man.txt
z88_man_template = """DYNAMIC START
---------------------------------------------------------------------------
Z88V14OS
---------------------------------------------------------------------------

---------------------------------------------------------------------------
GLOBAL
---------------------------------------------------------------------------

GLOBAL START
   IBFLAG          0
   IPFLAG          0
   IHFLAG          $z88_param_ihflag
GLOBAL END

---------------------------------------------------------------------------
LINEAR SOLVER
---------------------------------------------------------------------------

SOLVER START
   MAXIT           10000
   EPS             1e-007
   RALPHA          0.0001
   ROMEGA          1.1
SOLVER END

---------------------------------------------------------------------------
STRESS
---------------------------------------------------------------------------

STRESS START
   KDFLAG        0
   ISFLAG        $z88_param_isflag
STRESS END

DYNAMIC END
"""

# for memory parameter file z88.dyn
z88_dyn_template = """DYNAMIC START
---------------------------------------------------------------------------
Z88 new version 14OS                   Z88 neue Version 14OS
---------------------------------------------------------------------------

---------------------------------------------------------------------------
LANGUAGE                   SPRACHE
---------------------------------------------------------------------------
GERMAN

---------------------------------------------------------------------------
Entries for mesh generator Z88N        Daten fuer Netzgenerator
---------------------------------------------------------------------------
  NET START
    MAXSE  40000
    MAXESS   800
    MAXKSS  4000
    MAXAN     15
  NET END

---------------------------------------------------------------------------
Common entries for all modules         gemeinsame Daten fuer alle Module
---------------------------------------------------------------------------

  COMMON START
    MAXGS  100000000
    MAXKOI   2800000
    MAXK       60000
    MAXE      300000
    MAXNFG    200000
    MAXMAT        32
    MAXPEL        32
    MAXJNT        32
    MAXPR      10000
    MAXRBD     15000
    MAXIEZ   6000000
    MAXGP    2000000
  COMMON END

---------------------------------------------------------------------------
Entries for Cuthill-McKee Z88H         Daten fuer Cuthill- McKee Programm
---------------------------------------------------------------------------
  CUTKEE START
    MAXGRA  200
    MAXNDL 1000
  CUTKEE END


DYNAMIC END
"""

##  @}



# ***************************************************************************
# *   Copyright (c) 2018 looooo <sppedflyer@gmail.com>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

import os
import ast
import unittest
import FreeCAD as App


def test_python_syntax(rootdir, whitelist=None):
    whitelist = whitelist or []
    log = []
    for sub_dir, dirs, files in os.walk(rootdir):
        for fn in files:
            kargs = {}
            kargs["encoding"] = "utf-8"
            if (not fn in whitelist) and os.path.splitext(fn)[1] == ".py":
                with open(os.path.join(sub_dir, fn), **kargs) as py_file:
                    try:
                        ast.parse(py_file.read())
                    except SyntaxError as err:
                        log.append(str(err).replace("<unknown>", os.path.join(sub_dir, fn)))
    message = "\n\n" + "#" * 30 + "\n"
    message += "{} python files are not parseable:\n\n".format(len(log))
    for i, m in enumerate(log):
        message += str(i + 1) + " " + m + "\n"
    if log:
        raise RuntimeError(
            "there are some files not parse-able with the used python-interpreter" + message
        )
    else:
        return


class PythonSyntaxTestCase(unittest.TestCase):
    """
    Test Case to test python syntax of all python files in FreeCAD
    """

    def setUp(self):
        self.whitelist = []
        self.whitelist += [
            "ap203_configuration_controlled_3d_design_of_mechanical_parts_and_assemblies_mim_lf.py"
        ]
        self.whitelist += ["automotive_design.py"]
        self.whitelist += ["ifc2x3.py"]
        self.whitelist += ["ifc4.py"]

    def testAll(self):
        mod_dir = os.path.join(App.getHomePath(), "Mod")
        test_python_syntax(mod_dir, self.whitelist)



# SPDX-License-Identifier: LGPL-2.1-or-later

import FreeCAD
import FreeCADGui
from PySide import QtCore, QtGui


class DxfImportDialog:
    """
    A controller class that creates, manages, and shows the DXF import dialog.
    """

    def __init__(self, entity_counts, parent=None):
        # Step 1: Load the UI from the resource file. This returns a new QDialog instance.
        self.dialog = FreeCADGui.PySideUic.loadUi(":/ui/preferences-dxf-import.ui")

        # Now, all widgets like "label_Summary" are attributes of self.dialog

        self.entity_counts = entity_counts
        self.total_entities = sum(entity_counts.values())

        self.setup_ui()
        self.connect_signals()
        self.load_settings_and_set_initial_state()

    def setup_ui(self):
        """Perform initial UI setup."""
        self.dialog.label_Summary.setText(
            f"File contains approximately {self.total_entities} geometric entities."
        )
        self.dialog.label_Warning.hide()

    def connect_signals(self):
        """Connect signals from the dialog's widgets to our methods."""
        buttonBox = self.dialog.findChild(QtGui.QDialogButtonBox, "buttonBox")
        if buttonBox:
            # Connect to our custom slots INSTEAD of the dialog's built-in ones
            buttonBox.accepted.connect(self.on_accept)
            buttonBox.rejected.connect(self.on_reject)
            FreeCAD.Console.PrintLog("DxfImportDialog: OK and Cancel buttons connected.\n")
        else:
            FreeCAD.Console.PrintWarning("DxfImportDialog: Could not find buttonBox!\n")

        self.dialog.radio_ImportAs_Draft.toggled.connect(self.update_warning_label)
        self.dialog.radio_ImportAs_Primitives.toggled.connect(self.update_warning_label)
        self.dialog.radio_ImportAs_Shapes.toggled.connect(self.update_warning_label)
        self.dialog.radio_ImportAs_Fused.toggled.connect(self.update_warning_label)

    def on_accept(self):
        """Custom slot to debug the OK button click."""
        FreeCAD.Console.PrintLog(
            "DxfImportDialog: 'OK' button clicked. Calling self.dialog.accept().\n"
        )
        # Manually call the original slot
        self.dialog.accept()
        FreeCAD.Console.PrintLog("DxfImportDialog: self.dialog.accept() has been called.\n")

    def on_reject(self):
        """Custom slot to debug the Cancel button click."""
        FreeCAD.Console.PrintLog(
            "DxfImportDialog: 'Cancel' button clicked. Calling self.dialog.reject().\n"
        )
        # Manually call the original slot
        self.dialog.reject()
        FreeCAD.Console.PrintLog("DxfImportDialog: self.dialog.reject() has been called.\n")

    def load_settings_and_set_initial_state(self):
        """Load saved preferences and set the initial state of the dialog."""
        hGrp = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")

        mode = hGrp.GetInt("DxfImportMode", 2)

        if mode == 0:
            self.dialog.radio_ImportAs_Draft.setChecked(True)
        elif mode == 1:
            self.dialog.radio_ImportAs_Primitives.setChecked(True)
        elif mode == 3:
            self.dialog.radio_ImportAs_Fused.setChecked(True)
        else:
            self.dialog.radio_ImportAs_Shapes.setChecked(True)

        is_legacy = hGrp.GetBool("dxfUseLegacyImporter", False)
        if is_legacy:
            self.dialog.radio_ImportAs_Primitives.setEnabled(False)
            self.dialog.radio_ImportAs_Draft.setEnabled(True)
            self.dialog.radio_ImportAs_Shapes.setEnabled(True)
            self.dialog.radio_ImportAs_Fused.setEnabled(True)
        else:
            self.dialog.radio_ImportAs_Draft.setEnabled(True)
            self.dialog.radio_ImportAs_Primitives.setEnabled(True)
            self.dialog.radio_ImportAs_Shapes.setEnabled(True)
            self.dialog.radio_ImportAs_Fused.setEnabled(True)

        self.update_warning_label()

    def update_warning_label(self):
        """Updates the warning label based on selection and entity count."""
        self.dialog.label_Warning.hide()
        current_mode = self.get_selected_mode()

        if self.total_entities > 5000 and (current_mode == 0 or current_mode == 1):
            self.dialog.label_Warning.setText(
                "Warning: Importing over 5000 entities as editable objects can be very slow."
            )
            self.dialog.label_Warning.show()
        elif self.total_entities > 20000 and current_mode == 2:
            self.dialog.label_Warning.setText(
                "Warning: Importing over 20,000 entities as individual shapes may be slow."
            )
            self.dialog.label_Warning.show()

    def exec_(self):
        FreeCAD.Console.PrintLog("DxfImportDialog: Calling self.dialog.exec_()...\n")
        result = self.dialog.exec_()
        FreeCAD.Console.PrintLog(
            "DxfImportDialog: self.dialog.exec_() returned with result: {}\n".format(result)
        )
        # QDialog.Accepted is usually 1, Rejected is 0.
        FreeCAD.Console.PrintLog(
            "(Note: QDialog.Accepted = {}, QDialog.Rejected = {})\n".format(
                QtGui.QDialog.Accepted, QtGui.QDialog.Rejected
            )
        )
        return result

    def get_selected_mode(self):
        """Return the integer value of the selected import mode."""
        if self.dialog.radio_ImportAs_Draft.isChecked():
            return 0
        if self.dialog.radio_ImportAs_Primitives.isChecked():
            return 1
        if self.dialog.radio_ImportAs_Fused.isChecked():
            return 3
        if self.dialog.radio_ImportAs_Shapes.isChecked():
            return 2
        return 2

    def get_show_dialog_again(self):
        """Return True if the dialog should be shown next time."""
        return not self.dialog.checkBox_ShowDialogAgain.isChecked()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2025 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the task panel code for the Draft ShapeString tool."""
## @package task_shapestring
# \ingroup drafttaskpanels
# \brief Provides the task panel code for the Draft ShapeString tool.

## \addtogroup drafttaskpanels
# @{
import PySide.QtCore as QtCore
import PySide.QtGui as QtGui

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc
import WorkingPlane
from draftguitools import gui_tool_utils
from draftutils import params
from draftutils.translate import translate
from DraftVecUtils import toString

# So the resource file doesn't trigger errors from code checkers (flake8)
True if Draft_rc.__name__ else False


class ShapeStringTaskPanel:
    """Base class for Draft_ShapeString task panel."""

    def __init__(self, point=None, size=None, string="", font=""):

        self.form = Gui.PySideUic.loadUi(":/ui/TaskShapeString.ui")
        self.form.setWindowTitle(translate("draft", "ShapeString"))
        self.form.setWindowIcon(QtGui.QIcon(":/icons/Draft_ShapeString.svg"))

        self.global_mode = params.get_param("GlobalMode")
        self.form.cbGlobalMode.setChecked(self.global_mode)
        self.change_coord_labels()

        self.wp = WorkingPlane.get_working_plane()
        if point is not None:
            self.point = point
        elif self.global_mode:
            self.point = App.Vector()
        else:
            self.point = self.wp.position
        self.display_point(self.point)
        self.pointPicked = False

        self.height = size if size is not None else params.get_param("ShapeStringHeight")
        self.form.sbHeight.setProperty("rawValue", self.height)
        self.text = string if string else params.get_param("ShapeStringText")
        self.form.leText.setText(self.text)
        self.platform_win_dialog("Overwrite")
        self.font_file = font if font else params.get_param("ShapeStringFontFile")
        self.form.fcFontFile.setFileName(self.font_file)

        # Prevent cyclic processing of point values:
        self.display_point_active = False
        # Default for the "DontUseNativeFontDialog" preference:
        self.font_dialog_pref = False
        # Dummy attribute used by gui_tool_utils.getPoint in action method
        self.node = None

        self.form.sbX.valueChanged.connect(self.set_point_x)
        self.form.sbY.valueChanged.connect(self.set_point_y)
        self.form.sbZ.valueChanged.connect(self.set_point_z)
        if hasattr(self.form.cbGlobalMode, "checkStateChanged"):  # Qt version >= 6.7.0
            self.form.cbGlobalMode.checkStateChanged.connect(self.set_global_mode)
        else:  # Qt version < 6.7.0
            self.form.cbGlobalMode.stateChanged.connect(self.set_global_mode)
        self.form.pbReset.clicked.connect(self.reset_point)
        self.form.sbHeight.valueChanged.connect(self.set_height)
        self.form.leText.textEdited.connect(self.set_text)
        self.form.fcFontFile.fileNameSelected.connect(self.set_file)

    def action(self, arg):
        """scene event handler"""
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.reject()
        elif arg["Type"] == "SoLocation2Event":  # mouse movement detection
            if not self.pointPicked:
                self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg, noTracker=True)
                self.display_point(self.point)
                self.update_hints()
        elif arg["Type"] == "SoMouseButtonEvent":
            if (arg["State"] == "DOWN") and (arg["Button"] == "BUTTON1"):
                self.display_point(self.point)
                self.pointPicked = True
                self.update_hints()

    def change_coord_labels(self):
        if self.global_mode:
            self.form.labelX.setText(translate("draft", "Global {}").format("X"))
            self.form.labelY.setText(translate("draft", "Global {}").format("Y"))
            self.form.labelZ.setText(translate("draft", "Global {}").format("Z"))
        else:
            self.form.labelX.setText(translate("draft", "Local {}").format("X"))
            self.form.labelY.setText(translate("draft", "Local {}").format("Y"))
            self.form.labelZ.setText(translate("draft", "Local {}").format("Z"))

    def display_point(self, point):
        """Display the selected point."""
        self.display_point_active = True
        if not self.global_mode:
            point = self.wp.get_local_coords(point)
        self.form.sbX.setProperty("rawValue", point.x)
        self.form.sbY.setProperty("rawValue", point.y)
        self.form.sbZ.setProperty("rawValue", point.z)
        self.display_point_active = False

    def escape_string(self, string):
        return string.replace("\\", "\\\\").replace('"', '\\"')

    def platform_win_dialog(self, flag):
        """Handle the type of dialog depending on the platform."""
        ParamGroup = App.ParamGet("User parameter:BaseApp/Preferences/Dialog")
        if flag == "Overwrite":
            if "DontUseNativeFontDialog" not in ParamGroup.GetBools():
                # initialize nonexisting one
                ParamGroup.SetBool("DontUseNativeFontDialog", True)
            param = ParamGroup.GetBool("DontUseNativeFontDialog")
            self.font_dialog_pref = ParamGroup.GetBool("DontUseNativeDialog")
            ParamGroup.SetBool("DontUseNativeDialog", param)
        elif flag == "Restore":
            ParamGroup.SetBool("DontUseNativeDialog", self.font_dialog_pref)

    def reset_point(self):
        """Reset the selected point and display new point in the task panel."""
        if self.global_mode:
            self.point = App.Vector()
        else:
            self.point = self.wp.position
        self.pointPicked = False
        self.display_point(self.point)
        self.update_hints()

    def set_file(self, val):
        """Assign the selected font file."""
        self.font_file = val
        params.set_param("ShapeStringFontFile", self.font_file)

    def set_global_mode(self, val):
        self.global_mode = bool(getattr(val, "value", val))
        params.set_param("GlobalMode", self.global_mode)
        self.change_coord_labels()
        self.display_point(self.point)

    def set_height(self, val):
        self.height = App.Units.Quantity(val).Value
        params.set_param("ShapeStringHeight", self.height)

    def set_point_coord(self, coord, val):
        """Change self.point based on  X, Y or Z value entered in the task panel.

        coord should be "x", "y" or "z".
        """
        if self.display_point_active:
            return
        if self.global_mode:
            point = self.point
        else:
            point = self.wp.get_local_coords(self.point)
        setattr(point, coord, App.Units.Quantity(val).Value)
        if self.global_mode:
            self.point = point
        else:
            self.point = self.wp.get_global_coords(point)

    def set_point_x(self, val):
        self.set_point_coord("x", val)

    def set_point_y(self, val):
        self.set_point_coord("y", val)

    def set_point_z(self, val):
        self.set_point_coord("z", val)

    def set_text(self, val):
        self.text = val
        params.set_param("ShapeStringText", self.text)

    def update_hints(self):
        if self.pointPicked:
            Gui.HintManager.hide()
        else:
            Gui.HintManager.show(
                Gui.InputHint(translate("draft", "%1 pick point"), Gui.UserInput.MouseLeft)
            )


class ShapeStringTaskPanelCmd(ShapeStringTaskPanel):
    """Task panel for Draft_ShapeString."""

    def __init__(self, sourceCmd):
        super().__init__()
        self.sourceCmd = sourceCmd

    def accept(self):
        """Execute when clicking the OK button."""
        self.create_object()
        self.reject()
        return True

    def create_object(self):
        """Create object in the current document."""
        Gui.addModule("Draft")
        Gui.addModule("WorkingPlane")
        cmd = "Draft.make_shapestring("
        cmd += 'String="' + self.escape_string(self.text) + '", '
        cmd += 'FontFile="' + self.escape_string(self.font_file) + '", '
        cmd += "Size=" + str(self.height) + ", "
        cmd += "Tracking=0.0"
        cmd += ")"
        self.sourceCmd.commit(
            translate("draft", "Create ShapeString"),
            [
                "ss = " + cmd,
                "pl = FreeCAD.Placement()",
                "pl.Base = " + toString(self.point),
                "pl.Rotation = WorkingPlane.get_working_plane().get_placement().Rotation",
                "ss.Placement = pl",
                "Draft.autogroup(ss)",
                "FreeCAD.ActiveDocument.recompute()",
            ],
        )

    def reject(self):
        """Run when clicking the Cancel button."""
        self.sourceCmd.finish()
        self.platform_win_dialog("Restore")
        return True


class ShapeStringTaskPanelEdit(ShapeStringTaskPanel):
    """Task panel for Draft ShapeString object in edit mode."""

    def __init__(self, vobj):

        self.obj = vobj.Object
        super().__init__(
            self.obj.Placement.Base, self.obj.Size.Value, self.obj.String, self.obj.FontFile
        )
        self.pointPicked = True
        self.call = Gui.activeView().addEventCallback("SoEvent", self.action)

    def accept(self):

        Gui.doCommand('ss = FreeCAD.ActiveDocument.getObject("' + self.obj.Name + '")')
        Gui.doCommand('ss.String="' + self.escape_string(self.text) + '"')
        Gui.doCommand('ss.FontFile="' + self.escape_string(self.font_file) + '"')
        Gui.doCommand("ss.Size=" + str(self.height))
        Gui.doCommand("ss.Placement.Base=" + toString(self.point))
        Gui.doCommand("FreeCAD.ActiveDocument.recompute()")

        self.reject()
        return True

    def finish(self):

        Gui.activeView().removeEventCallback("SoEvent", self.call)
        Gui.Snapper.off()
        Gui.Control.closeDialog()
        return None

    def reject(self):

        self.obj.ViewObject.Document.resetEdit()
        self.platform_win_dialog("Restore")
        return True


## @}
